"""spec2nii module containing functions specific to interpreting Philips data/list
format data. Must be paired with a SPAR dataset as well.
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford
"""
import re

import pandas as pd
import numpy as np
from nibabel.nicom.dicomwrappers import wrapper_from_file
from pydicom.errors import InvalidDicomError

from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext

from spec2nii.Philips.philips import read_spar, spar_to_nmrs_hdrext, _philips_orientation
from spec2nii.Philips.philips_dcm import \
    _enhanced_dcm_svs_to_orientation, \
    _is_new_format, \
    _extractDicomMetadata_new, \
    _extractDicomMetadata_old
from spec2nii.nifti_orientation import NIFTIOrient


index_options = ['STD', 'REJ', 'PHC', 'FRC', 'NOI', 'NAV']
index_headers = ['typ', 'mix', 'dyn', 'card', 'echo', 'loca',
                 'chan', 'extr1', 'extr2', 'kx', 'ky', 'kz',
                 'aver', 'sign', 'rf', 'grad', 'enc', 'rtop',
                 'rr', 'size', 'offset']
indices = ['chan', 'aver', 'dyn', 'card', 'echo', 'loca', 'extr1', 'extr2']

defaults = {'chan': 'DIM_COIL',
            'aver': 'DIM_DYN',
            'dyn': 'DIM_DYN',
            'echo': 'DIM_INDIRECT_0'}


class TooManyDimError(Exception):
    pass


def read_data_list_pair(data_file, list_file, aux_file, special_case=None):

    df, num_dict, coord_dict, os_dict = _read_list(list_file)
    sorted_data_dict = _read_data(data_file, df)

    if aux_file.suffix.lower() == '.spar':
        spar_params = read_spar(aux_file)

        dwelltime = 1.0 / float(spar_params["sample_frequency"])
        orientation = NIFTIOrient(_philips_orientation(spar_params))

        # Meta
        base_meta = spar_to_nmrs_hdrext(spar_params)
    else:
        try:
            dcm_obj = wrapper_from_file(aux_file)
        except (InvalidDicomError, FileNotFoundError) as exc:
            raise ValueError('Data/List auxiliary file must be a valid Philips .SPAR or DICOM file.') from exc

        orientation = _enhanced_dcm_svs_to_orientation(dcm_obj)

        if _is_new_format(dcm_obj):
            base_meta = _extractDicomMetadata_new(dcm_obj)
            dwelltime = 1.0 / dcm_obj.dcm_data.SpectralWidth
        else:
            base_meta = _extractDicomMetadata_old(dcm_obj)
            dwelltime = 1.0 / dcm_obj.dcm_data[('2005', '1357')].value

    base_meta.set_standard_def(
        'OriginalFile',
        [data_file.name,
         list_file.name,
         aux_file.name])

    data_out = []
    name_out = []
    for data_type in sorted_data_dict:
        data = sorted_data_dict[data_type]
        name = data_type

        meta = base_meta.copy()

        kept_ind = []
        for ii, sha in zip(indices, data.shape[1:]):
            if sha > 1:
                kept_ind.append(ii)

        out_data = data.squeeze()
        if len(kept_ind) > 3:
            raise TooManyDimError('Number of additional dimensions > 3.'
                                  f' Dimensions are {kept_ind} with shape {out_data.shape[1:]}.'
                                  ' NIFTI-MRS can only handle three dynamic dimensions. Unsure how to proceed.')

        # Special cases
        if data_type == 'STD_0'\
                and (special_case == 'hyper' or 'hyper' in meta['ProtocolName'].lower()):
            out_hyper, meta_hyper = _special_case_hyper(out_data, meta)
            # Handle the main acquisition of the HYPER (short TE + editing) sequence

            # Insert spatial dimensions
            out_shortte = out_hyper[0].reshape((1, 1, 1) + out_hyper[0].shape)
            out_edited = out_hyper[1].reshape((1, 1, 1) + out_hyper[1].shape)

            # Apply conjugate
            out_shortte = out_shortte.conj()
            out_edited = out_edited.conj()

            data_out.append(
                gen_nifti_mrs_hdr_ext(out_shortte, dwelltime, meta_hyper[0], orientation.Q44, no_conj=True))
            name_out.append('hyper_short_te')

            data_out.append(
                gen_nifti_mrs_hdr_ext(out_edited, dwelltime, meta_hyper[1], orientation.Q44, no_conj=True))
            name_out.append('hyper_edited')

            continue

        elif data_type == 'STD_1'\
                and (special_case == 'hyper' or 'hyper' in meta['ProtocolName'].lower()):
            # Handle the water ref acquisition of the HYPER sequence

            meta.set_dim_info(
                0,
                'DIM_COIL')

            meta.set_dim_info(
                1,
                'DIM_USER_0',
                info='HYPER water reference')

            name = 'hyper_water_ref'

        else:
            # Default case: assign default DIM Tags.
            unknown_counter = 0
            for idx, ki in enumerate(kept_ind):
                if ki in defaults:
                    meta.set_dim_info(
                        idx,
                        defaults[ki],
                        info=f'data/list dim {ki}')
                else:
                    meta.set_dim_info(
                        idx,
                        f'DIM_USER_{unknown_counter}',
                        info=f'data/list dim {ki}')
                    unknown_counter += 1

        # Insert spatial dimensions
        out_data = out_data.reshape((1, 1, 1) + out_data.shape)
        # Apply conjugate
        out_data = out_data.conj()

        data_out.append(gen_nifti_mrs_hdr_ext(out_data, dwelltime, meta, orientation.Q44, no_conj=True))
        name_out.append(name)

    return data_out, name_out


def _read_data(data_file, df):
    '''Read .data file'''
    with open(data_file, 'rb') as f:
        raw = np.fromfile(f, dtype='<f4')
        raw = raw[0::2] + 1j * raw[1::2]

    data_types = df.typ.unique()
    output_dict = {}
    for tt in data_types:
        curr_df = df.loc[df.typ == tt, :]
        if tt == 'NOI':
            # Special simple case for NOI
            spec_res = int(curr_df['size'].max() / 8)
            ncha = curr_df['chan'].max() + 1
            nloca = curr_df['loca'].max() + 1
            output_dict[tt] = np.zeros((spec_res, ncha, nloca), dtype=complex)
        else:
            n_mix = curr_df['mix'].max() + 1
            # Other types might use all the loops
            for mix in range(n_mix):
                curr_mix_df = curr_df.loc[curr_df.mix == mix, :]
                shape = []
                shape.append(int(curr_mix_df['size'].max() / 8))
                for ind in indices:
                    shape.append(curr_mix_df[ind].max() + 1)
                output_dict[f'{tt}_{mix}'] = np.zeros(shape, dtype=complex)

    # Now extract data
    offset = 0
    for ind in range(df.shape[0]):
        cdf = df.iloc[ind]
        tt = cdf.typ
        dsize = int(cdf['size'].max() / 8)
        if tt == 'NOI':
            output_dict[tt][:, cdf.chan, cdf.loca] = raw[offset:(offset + dsize)]
        else:
            mix = cdf.mix
            ind = [cdf[ii] for ii in indices]
            ind = tuple([slice(None), ] + ind)
            output_dict[f'{tt}_{mix}'][ind] = raw[offset:(offset + dsize)]
        offset += dsize

    # Slight hack - remove all zero data if there is any (skipped indices)
    for key in output_dict:
        if key == 'NOI':
            continue
        keyparts = key.split('_')
        cdf = df[df.typ == keyparts[0]][df.mix == int(keyparts[1])]
        used_df = cdf[indices].loc[:, cdf[indices].max() > 0]
        used_col_indices = [indices.index(col) for col in used_df.columns]
        for idx, col_idx in enumerate(used_col_indices):
            used_indices_in_col = used_df.iloc[:, idx].unique()
            output_dict[key] = np.take(output_dict[key], used_indices_in_col, axis=col_idx + 1)

    return output_dict


def _read_list(list_file):
    '''Read .list file'''
    num_dict = {}
    coord_dict = {}
    os_dict = {}
    index_array = []

    n_mixes = 1
    n_echoes = 1
    with open(list_file, 'r') as lf:
        for line in lf:
            line = line.strip()
            if not line:
                continue
            elif line[0] == '#':
                continue
            else:
                if 'number_of_mixes' in line:
                    n_mixes = int(line[-4:])
                elif 'number_of_echoes' in line:
                    n_echoes = int(line[-4:])
                elif 'number_of_' in line:
                    matched = re.search(
                        r'(\d+)\s+\d+\s+\d+\s+number_of_([a-z0-9_]+)\s+:\s+(\d+)',
                        line)
                    if matched[2] not in num_dict:
                        num_dict[matched[2]] = np.zeros((n_mixes), dtype=int)
                    num_dict[matched[2]][int(matched[1])] = int(matched[3])
                elif '_range' in line:
                    matched = re.search(
                        r'(\d+)\s+(\d+)\s+\d+\s+([a-z]+_range)\s+:\s+(-?\d+)\s+(-?\d+)',
                        line,
                        flags=re.IGNORECASE)
                    if matched[3] not in coord_dict:
                        coord_dict[matched[3]] = np.zeros((n_mixes, n_echoes, 2), dtype=int)
                    coord_dict[matched[3]][int(matched[1]), int(matched[2]), 0] = int(matched[4])
                    coord_dict[matched[3]][int(matched[1]), int(matched[2]), 1] = int(matched[5])
                elif '_oversample_factor' in line:
                    matched = re.search(
                        r'(\d+)\s+(\d+)\s+\d+\s+([a-z]+_oversample_factor)\s+:\s+(\d+)',
                        line,
                        flags=re.IGNORECASE)
                    if matched[3] not in os_dict:
                        os_dict[matched[3]] = np.zeros((n_mixes, n_echoes), dtype=float)
                    os_dict[matched[3]][int(matched[1]), int(matched[2])] = int(matched[4])
                elif any(opt in line for opt in index_options):
                    index_array.append(pd.Series(line.split(), index=index_headers))

    df = pd.concat(index_array, axis=1).T
    df[index_headers[1:]] = df[index_headers[1:]].apply(pd.to_numeric)

    return df, num_dict, coord_dict, os_dict


# Special cases
def _special_case_hyper(data, meta):
    '''
    Notes:
      - (AG 03/22/2023) Updated to handle incomplete data.
    '''

    data_short_te = data[:, :, :32]
    data_edited = data[:, :, 32:]

    if data_edited.shape[-1] % 4 != 0:                                                      # Handle Incomplete Data
        old_num_avgs = data_edited.shape[-1]                                                # Old Number of Averages
        new_num_avgs = (data_edited.shape[-1] // 4) * 4                                     # Complete Sets of 4
        data_edited = data_edited[..., :new_num_avgs]                                       # Only Keep Complete Sets
        print(f'Correcting - Incomplete Averages {old_num_avgs} --> {new_num_avgs}    Corrected**')

    data_edited = data_edited.T.reshape((-1, 4, data.shape[1], data.shape[0])).T

    meta_short_te = meta.copy()
    meta_edited = meta.copy()

    meta_short_te.set_dim_info(
        0,
        'DIM_COIL')
    meta_short_te.set_dim_info(
        1,
        'DIM_DYN')

    edit_pulse_1   = 1.9
    edit_pulse_2   = 4.58
    edit_pulse_off = 4.18
    dim_info       = "HERCULES j-difference editing, four conditions"
    dim_header     = {"EditCondition": ["A", "B", "C", "D"]}
    edit_pulse_val = {
        "A": {"PulseOffset": [edit_pulse_1, edit_pulse_2], "PulseDuration": 0.02},
        "B": {"PulseOffset": [edit_pulse_off, edit_pulse_2], "PulseDuration": 0.02},
        "C": {"PulseOffset": edit_pulse_1, "PulseDuration": 0.02},
        "D": {"PulseOffset": edit_pulse_off, "PulseDuration": 0.02}}

    meta_edited.set_dim_info(
        0,
        'DIM_COIL')

    meta_edited.set_dim_info(
        1,
        'DIM_EDIT',
        dim_info,
        dim_header)

    meta_edited.set_dim_info(2, 'DIM_DYN')
    meta_edited.set_standard_def("EditPulse", edit_pulse_val)

    return [data_short_te, data_edited], \
           [meta_short_te, meta_edited]



'''Tests for GE pfile format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs
from .io_for_tests import read_nifti_mrs_with_hdr
from nifti_mrs.nifti_mrs import NIFTI_MRS

# Data paths
ge_path = Path(__file__).parent / 'spec2nii_test_data' / 'ge'
svs_path = ge_path / 'pFiles' / 'svMRS' / 'P03072.7'
mrsi_path = ge_path / 'pFiles' / 'MRSI' / 'P18432.7'
mpress_path = ge_path / 'pFiles' / 'big_gaba' / 'S01_GABA_68.7'

# Test set from Mark Mikkelsen
mm_herc = [
    ge_path / 'pFiles' / 'HERCULES' / 'sub-01_ses-01_acq-herculespress_svs_noID.7',
    ge_path / 'pFiles' / 'HERCULES' / 'sub-01_ses-01_acq-herculesslaser_svs_noID.7']
mm_hermes = [
    ge_path / 'pFiles' / 'HERMES' / 'sub-01_ses-01_acq-hermespress_svs_noID.7',
    ge_path / 'pFiles' / 'HERMES' / 'sub-01_ses-01_acq-hermesslaser_svs_noID.7']
mm_mega = ge_path / 'pFiles' / 'MEGA' / 'Big_GABA'
mm_press = ge_path / 'pFiles' / 'PRESS' / 'Big_GABA'
mm_press_noid = ge_path / 'pFiles' / 'PRESS' / 'sub-01_ses-01_acq-press_svs_noID.7'
mm_slaser = ge_path / 'pFiles' / 'sLASER' / 'sub-01_ses-01_acq-slaser_svs_noID.7'

# HBCD / ISTHMUS datasets
hbcd2_path = ge_path / 'pFiles' / 'hbcd' / 'P31744.7'


def test_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'ge',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           str(svs_path)])

    img = read_nifti_mrs(tmp_path / 'svs.nii.gz')
    img_ref = read_nifti_mrs(tmp_path / 'svs_ref.nii.gz')

    assert img.shape == (1, 1, 1, 4096, 32, 2)
    assert np.iscomplexobj(img.dataobj)
    assert 1 / img.header['pixdim'][4] == 5000.0

    assert img_ref.shape == (1, 1, 1, 4096, 32, 2)
    assert np.iscomplexobj(img_ref.dataobj)
    assert 1 / img_ref.header['pixdim'][4] == 5000.0

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'
    assert hdr_ext['SpectrometerFrequency'][0] == 127.76365
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == svs_path.name

    assert np.isclose(hdr_ext['EchoTime'], 0.27)
    assert np.isclose(hdr_ext['RepetitionTime'], 2.0)


def test_mrsi(tmp_path):

    subprocess.check_call(['spec2nii', 'ge',
                           '-f', 'mrsi',
                           '-o', tmp_path,
                           '-j',
                           str(mrsi_path)])

    img = read_nifti_mrs(tmp_path / 'mrsi.nii.gz')

    assert img.shape == (8, 8, 1, 1024, 32)
    assert np.iscomplexobj(img.dataobj)
    assert 1 / img.header['pixdim'][4] == 4000.0

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.763607
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == mrsi_path.name


def test_mpress(tmp_path):

    subprocess.check_call(['spec2nii', 'ge',
                           '-f', 'mpress',
                           '-o', tmp_path,
                           '-j',
                           str(mpress_path)])

    img = read_nifti_mrs(tmp_path / 'mpress.nii.gz')
    img_ref = read_nifti_mrs(tmp_path / 'mpress_ref.nii.gz')

    assert img.shape == (1, 1, 1, 2048, 8, 160, 2)
    assert np.iscomplexobj(img.dataobj)
    assert 1 / img.header['pixdim'][4] == 2000.0

    assert img_ref.shape == (1, 1, 1, 2048, 8, 8, 2)
    assert np.iscomplexobj(img_ref.dataobj)
    assert 1 / img_ref.header['pixdim'][4] == 2000.0

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'
    assert hdr_ext['dim_7'] == 'DIM_EDIT'
    assert hdr_ext['SpectrometerFrequency'][0] == 127.758139
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == mpress_path.name

    assert np.isclose(hdr_ext['EchoTime'], 0.068)
    assert np.isclose(hdr_ext['RepetitionTime'], 2.0)


def test_mm_hercules(tmp_path):
    for path in mm_herc:
        subprocess.run([
            'spec2nii', 'ge',
            '-f', path.stem,
            '-o', tmp_path,
            '-j',
            path])

        img, hdr_ext = read_nifti_mrs_with_hdr(
            tmp_path / path.with_suffix('.nii.gz').name)

        assert img.shape == (1, 1, 1, 4096, 32, 56, 4)
        assert np.iscomplexobj(img.dataobj)
        assert 1 / img.header['pixdim'][4] == 5000.0
        assert hdr_ext['dim_5'] == 'DIM_COIL'
        assert hdr_ext['dim_6'] == 'DIM_DYN'
        assert hdr_ext['dim_7'] == 'DIM_EDIT'
        assert np.isclose(127.771, hdr_ext['SpectrometerFrequency'][0], atol=1E-3)
        assert hdr_ext['ResonantNucleus'][0] == '1H'
        assert hdr_ext['OriginalFile'][0] == path.name
        assert hdr_ext['WaterSuppressed']

        # Test that the MEGA editing header info is not added
        assert 'dim_7_header' not in hdr_ext
        assert 'EditPulse' not in hdr_ext

        img_ref, hdr_ext_ref = read_nifti_mrs_with_hdr(
            tmp_path / f'{path.stem}_ref.nii.gz')
        assert img_ref.shape[:5] == (1, 1, 1, 4096, 32)
        assert img_ref.shape[6] == 4
        assert not hdr_ext_ref['WaterSuppressed']
        assert hdr_ext_ref['dim_5'] == 'DIM_COIL'
        assert hdr_ext_ref['dim_6'] == 'DIM_DYN'
        assert hdr_ext_ref['dim_7'] == 'DIM_EDIT'


def test_mm_hermes(tmp_path):
    for path in mm_hermes:
        subprocess.run([
            'spec2nii', 'ge',
            '-f', path.stem,
            '-o', tmp_path,
            '-j',
            path])

        img, hdr_ext = read_nifti_mrs_with_hdr(
            tmp_path / path.with_suffix('.nii.gz').name)

        assert img.shape == (1, 1, 1, 4096, 32, 56, 4)
        assert np.iscomplexobj(img.dataobj)
        assert 1 / img.header['pixdim'][4] == 5000.0
        assert hdr_ext['dim_5'] == 'DIM_COIL'
        assert hdr_ext['dim_6'] == 'DIM_DYN'
        assert hdr_ext['dim_7'] == 'DIM_EDIT'
        assert np.isclose(127.771, hdr_ext['SpectrometerFrequency'][0], atol=1E-3)
        assert hdr_ext['ResonantNucleus'][0] == '1H'
        assert hdr_ext['OriginalFile'][0] == path.name
        assert hdr_ext['WaterSuppressed']

        assert 'dim_7_header' not in hdr_ext
        assert 'EditPulse' not in hdr_ext

        img_ref, hdr_ext_ref = read_nifti_mrs_with_hdr(
            tmp_path / f'{path.stem}_ref.nii.gz')
        assert img_ref.shape[:5] == (1, 1, 1, 4096, 32)
        assert img_ref.shape[6] == 4
        assert not hdr_ext_ref['WaterSuppressed']
        assert hdr_ext_ref['dim_5'] == 'DIM_COIL'
        assert hdr_ext_ref['dim_6'] == 'DIM_DYN'
        assert hdr_ext_ref['dim_7'] == 'DIM_EDIT'


def test_mm_mega(tmp_path):
    for path in mm_mega.rglob('*.7'):
        subprocess.run([
            'spec2nii', 'ge',
            '-f', path.stem,
            '-o', tmp_path,
            '-j',
            path])

        img, hdr_ext = read_nifti_mrs_with_hdr(
            tmp_path / path.with_suffix('.nii.gz').name)

        assert img.shape[:3] == (1, 1, 1)
        assert img.shape[3] in (2048, 4096)
        assert img.shape[6] == 2
        assert np.iscomplexobj(img.dataobj)
        assert 1 / img.header['pixdim'][4] in (2000.0, 5000.0)
        assert hdr_ext['dim_5'] == 'DIM_COIL'
        assert hdr_ext['dim_6'] == 'DIM_DYN'
        assert hdr_ext['dim_7'] == 'DIM_EDIT'
        assert np.isclose(127.7, hdr_ext['SpectrometerFrequency'][0], atol=1E-1)
        assert hdr_ext['ResonantNucleus'][0] == '1H'
        assert hdr_ext['OriginalFile'][0] == path.name
        assert hdr_ext['WaterSuppressed']

        assert 'dim_7_header' in hdr_ext
        assert 'EditPulse' in hdr_ext
        assert 'dim_7_info' in hdr_ext

        img_ref, hdr_ext_ref = read_nifti_mrs_with_hdr(
            tmp_path / f'{path.stem}_ref.nii.gz')
        assert img_ref.shape[:3] == (1, 1, 1)
        assert img_ref.shape[6] == 2
        # Check reference has same number of points and coils as main without checking number of coils
        assert img_ref.shape[3] == img.shape[3]
        assert img_ref.shape[4] == img.shape[4]
        assert not hdr_ext_ref['WaterSuppressed']
        assert hdr_ext_ref['dim_5'] == 'DIM_COIL'
        assert hdr_ext_ref['dim_6'] == 'DIM_DYN'
        assert hdr_ext_ref['dim_7'] == 'DIM_EDIT'


def test_mm_press(tmp_path):
    for path in mm_press.rglob('*.7'):
        subprocess.run([
            'spec2nii', 'ge',
            '-f', path.stem,
            '-o', tmp_path,
            '-j',
            path])

        img, hdr_ext = read_nifti_mrs_with_hdr(
            tmp_path / path.with_suffix('.nii.gz').name)

        assert img.shape[:3] == (1, 1, 1)
        assert img.shape[3] in (2048, 4096)
        assert np.iscomplexobj(img.dataobj)
        assert 1 / img.header['pixdim'][4] in (2000.0, 5000.0)
        assert hdr_ext['dim_5'] == 'DIM_COIL'
        assert hdr_ext['dim_6'] == 'DIM_DYN'

        assert np.isclose(127.7, hdr_ext['SpectrometerFrequency'][0], atol=1E-1)
        assert hdr_ext['ResonantNucleus'][0] == '1H'
        assert hdr_ext['OriginalFile'][0] == path.name
        assert hdr_ext['WaterSuppressed']

        img_ref, hdr_ext_ref = read_nifti_mrs_with_hdr(
            tmp_path / f'{path.stem}_ref.nii.gz')
        assert img_ref.shape[:3] == (1, 1, 1)
        # Check reference has same number of points and coils as main without checking actual number
        assert img_ref.shape[3] == img.shape[3]
        assert img_ref.shape[4] == img.shape[4]
        assert not hdr_ext_ref['WaterSuppressed']
        assert hdr_ext_ref['dim_5'] == 'DIM_COIL'
        assert hdr_ext_ref['dim_6'] == 'DIM_DYN'


def test_mm_press_noid(tmp_path):
    subprocess.run([
        'spec2nii', 'ge',
        '-f', mm_press_noid.stem,
        '-o', tmp_path,
        '-j',
        mm_press_noid])

    img, hdr_ext = read_nifti_mrs_with_hdr(
        tmp_path / mm_press_noid.with_suffix('.nii.gz').name)

    assert img.shape == (1, 1, 1, 4096, 32, 64)
    assert np.iscomplexobj(img.dataobj)
    assert 1 / img.header['pixdim'][4] == 5000.0
    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'
    assert np.isclose(127.771, hdr_ext['SpectrometerFrequency'][0], atol=1E-3)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == mm_press_noid.name

    img_ref, hdr_ext_ref = read_nifti_mrs_with_hdr(
        tmp_path / f'{mm_press_noid.stem}_ref.nii.gz')
    assert img_ref.shape[:5] == (1, 1, 1, 4096, 32)
    assert not hdr_ext_ref['WaterSuppressed']
    assert hdr_ext_ref['dim_5'] == 'DIM_COIL'


def test_mm_slaser(tmp_path):
    subprocess.run([
        'spec2nii', 'ge',
        '-f', mm_slaser.stem,
        '-o', tmp_path,
        '-j',
        mm_slaser])

    img, hdr_ext = read_nifti_mrs_with_hdr(
        tmp_path / mm_slaser.with_suffix('.nii.gz').name)

    assert img.shape == (1, 1, 1, 4096, 32, 64)
    assert np.iscomplexobj(img.dataobj)
    assert 1 / img.header['pixdim'][4] == 5000.0
    assert hdr_ext['dim_5'] == 'DIM_COIL'
    assert hdr_ext['dim_6'] == 'DIM_DYN'
    assert np.isclose(127.771, hdr_ext['SpectrometerFrequency'][0], atol=1E-3)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == mm_slaser.name

    img_ref, hdr_ext_ref = read_nifti_mrs_with_hdr(
        tmp_path / f'{mm_slaser.stem}_ref.nii.gz')
    assert img_ref.shape[:5] == (1, 1, 1, 4096, 32)
    assert not hdr_ext_ref['WaterSuppressed']
    assert hdr_ext_ref['dim_5'] == 'DIM_COIL'


def test_hbcd_isthmus(tmp_path):
    subprocess.run([
        'spec2nii', 'ge',
        '-f', 'hbcd',
        '-o', tmp_path,
        '-j',
        hbcd2_path])

    assert (tmp_path / 'hbcd_edited.nii.gz').is_file()
    assert (tmp_path / 'hbcd_ref_edited.nii.gz').is_file()
    assert (tmp_path / 'hbcd_ref_short_te.nii.gz').is_file()
    assert (tmp_path / 'hbcd_short_te.nii.gz').is_file()

    img = NIFTI_MRS(tmp_path / 'hbcd_edited.nii.gz')
    assert img.shape == (1, 1, 1, 2048, 56, 8, 4)
    assert img.dim_tags == ['DIM_DYN', 'DIM_COIL', 'DIM_EDIT']

    img = NIFTI_MRS(tmp_path / 'hbcd_ref_edited.nii.gz')
    assert img.shape == (1, 1, 1, 2048, 4, 8)
    assert img.dim_tags == ['DIM_DYN', 'DIM_COIL', None]

    img = NIFTI_MRS(tmp_path / 'hbcd_ref_short_te.nii.gz')
    assert img.shape == (1, 1, 1, 2048, 4, 8)
    assert img.dim_tags == ['DIM_DYN', 'DIM_COIL', None]

    img = NIFTI_MRS(tmp_path / 'hbcd_short_te.nii.gz')
    assert img.shape == (1, 1, 1, 2048, 32, 8)
    assert img.dim_tags == ['DIM_DYN', 'DIM_COIL', None]



"""Reader for GE p-files.

This code is taken from the VESPA project https://scion.duhs.duke.edu/vespa/project.
I therefore include their BSD statement here.

    Copyright (c) 2010, Duke University. All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:
        * Redistributions of source code must retain the above copyright notice,
          this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
        * Neither the name of Duke University nor the United States Department
          of Veterans Affairs may be used to endorse or promote products derived
          from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ''AS
    IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
    OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
    DAMAGE.

    For portions of this code, copyright and license information differs from
    the above. In these cases, copyright and/or license information is inline.

In turn some of the code in this file was derived from the Python package
pfile-tools project, https://github.com/njvack/pfile-tools
and as such we have included their BSD statement in this file.

    Copyright (c) 2012, Board of Regents of the University of Wisconsin
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list
          of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this
          list of conditions and the following disclaimer in the documentation and/or
          other materials provided with the distribution.
        * Neither the name of the University of Wisconsin nor the names of its
          contributors may be used to endorse or promote products derived from this
          software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
    ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
# Python modules
import math
import sys
import csv
import ctypes
import ctypes as ct
from collections import namedtuple

# Third party packages
import numpy as np

# Our Modules
from spec2nii.GE.ge_hdr_fields import get_pfile_hdr_fields

StructInfo = namedtuple("StructInfo", ["label", "depth", "value", "field_type", "size", "offset"])


class UnknownPfile(RuntimeError):
    pass


class RevisionNumLittle(ct.LittleEndianStructure):

    @property
    def major(self):
        return int(self.revision)

    _pack_   = 1
    _fields_ = [('revision', ct.c_float)]


class RevisionNumBig(ct.BigEndianStructure):

    @property
    def major(self):
        return int(self.revision)

    _pack_   = 1
    _fields_ = [('revision', ct.c_float)]


class Pfile:
    """
    This class was based on the style of code from the pfile-tools
    package written by Nathan Vack in that we use ctypes to organize
    the reading of binary structures into a Python readable class
    instance. We have also incorporated code from their struct
    utilities modules as part of our class in order to dump out the
    header information to the stdout or as a list of strings.

    We use a subset of header variables that are sufficient to read the
    data from P-files in which we are interested. The structure for these
    variables was adapted from similar code found in the UCSF Sivic project.


    """
    def __init__(self, fname):

        self.file_name  = fname
        self.version    = 0
        self.hdr        = None
        self.map        = None
        self.endian     = 'little'  # def for version >= 11

        self.read_header()

        if not self.is_ge_file:
            raise UnknownPfile("Not a known GE Pfile - fname = %s" % fname)

        self.map_data()

    # Properties --------------------------------------------------

    @property
    def is_ge_file(self):
        if self.version < 12:
            if "GE" in self.hdr.rhr_rh_logo:
                return True
            else:
                return False
        else:
            offset = self.hdr.rhr_rdb_hdr_off_data
            # ARC 20220209: coverage here is likely to be incomplete. This check
            # may well be redundant anyway, since revision number has
            # previously been validated in _version
            if offset in \
                    (
                        61464,   # bjs from Matlab script for ver = 9
                        66072,
                        145908,
                        149788,
                        150336,
                        157276,  # v24 empirical
                        213684,  # 26.002
                        219828,  # 27.000
                        228020   # 28.003,29.1,30.0
                    ):
                return True
            else:
                return False

    @property
    def is_svs(self):

        if self.map is None:
            return False
        else:
            return self.map.is_svs

    @property
    def get_mapper(self):

        if self.hdr is None:
            return None

        psd = self.hdr.rhi_psdname.decode('utf-8').lower()
        proto = self.hdr.rhs_se_desc.decode('utf-8').lower()
        if psd == 'hbcd' and "press" in proto:
            psd = self.hdr.rhs_se_desc.decode('utf-8').lower()

        # MM: Some 'gaba' psd strings contain full path names, so truncate to the end of the path
        if psd.endswith('gaba'):
            psd = 'gaba'

        numecho = self.hdr.rhi_numecho

        if psd in \
                (
                    'probe-p',
                    'probe-s',
                    'probe-p_ach',       # MM - added for Calgary PROBE-P sequence
                    'presscsi',
                    'fidcsi',            # bjs - added for Pom's fidcsi 13C data
                    'ia/stable/fidcsi',  # bjs - added for Kearny's 13C data
                    'presscsi_nfl',      # bjs - added for Govind's SVS data off v25
                    'epsi_3d_24',        # bjs - added for soher check of MIDAS Browndyke data
                    'fidall'             # WTC - added for JG's Hyperpolarised 13C data
                    'mrs-press'          # ATG - Added for GE ShortTE in hbcd
                ):
            mapper = PfileMapper
        elif psd in ('oslaser', 'slaser_cni', 'slaser') and numecho == 1:
            mapper = PfileMapperSlaser  # MM: If non-edited data, use PfileMapperSlaser
        elif psd == 'oslaser' and numecho > 1:
            mapper = PfileMapperGaba  # MM: If edited data, use PfileMapperGaba
        elif psd in \
                (
                    'jpress',            # wtc - added for J-edited data.
                    'jpress_ac',         # ARC - added for Bergen jpress patch
                    'gaba',              # wtc - added for Nottingham MEGA-PRESS sequence
                    'hbcd',              # ATG - added HBCD  - reuse GABA mapper
                    'hbcd1',             # ATG - added HBCD1 - reuse GABA mapper
                    'hbcd2',             # ATG - added HBCD2 - reuse GABA mapper
                    'probe-p-mega_rml',  # MM - added for Calgary MEGA-PRESS sequence
                    'repress7'           # MM - added for old CUBRIC data
                ):
            mapper = PfileMapperGaba
        elif psd == 'probe-sl':
            # wtc - added for CSI sequence from Manchester
            mapper = PfileMapperProbeSL
        else:
            raise UnknownPfile("No Pfile mapper for pulse sequence = %s" % psd)

        return mapper

    def read_header(self):

        filelike = open(self.file_name, 'rb')

        # determine version number of this header from revision of rdbm.h
        version = self._version(filelike)
        if version == 0:
            raise UnknownPfile("Pfile not supported for version %s" % version)

        # Here we dynamically configure the ctypes structures into which the
        # binary file will be read, based on the revision number
        #
        # Note. Determined empirically that I cannot declare the XxxHeader
        # class at the top level of the module with an attribute ._fields_ = []
        # and then append into it. I have to create a list and then assign
        # _fields_ attribute to that list in one step.  Don't know why.
        #
        # Note 2. Had to move Class definition into this function so that the
        # class can be reconstituted more than once for multiple GE file reads.
        # At the top level of the module, the _fields_ attribute could be
        # created once dynamically, but afterwards would stick around and
        # could not then be changed.

        if self.endian == 'big':
            class PfileHeaderBig(ct.BigEndianStructure):
                """
                Contains the ctypes Structure for a GE P-file rdb header.
                Dynamically allocate the ctypes _fields_ list later depending on revision
                """
                _pack_   = 1
                _fields_ = get_pfile_hdr_fields(version)
                # _fields_ = utilge.get_pfile_hdr_fields(version)
            hdr = PfileHeaderBig()
        else:
            class PfileHeaderLittle(ct.LittleEndianStructure):
                """
                Contains the ctypes Structure for a GE P-file rdb header.
                Dynamically allocate the ctypes _fields_ list later depending on revision
                """
                _pack_   = 1
                _fields_ = get_pfile_hdr_fields(version)
                # _fields_ = utilge.get_pfile_hdr_fields(version)
            hdr = PfileHeaderLittle()

        try:
            # read  header information from start of file
            filelike.seek(0)
            filelike.readinto(hdr)
            filelike.close()
        except Exception:
            filelike.close()
            raise UnknownPfile("Trouble reading file into header structure for version %s" % version)

        self.version = version
        self.hdr = hdr

    def map_data(self):
        """
        Select appropriate mapper class using the pulse sequence name string,
        instantiate and read the data from the file into the 'map' attribute

        """
        mapper = self.get_mapper
        self.map = mapper(self.file_name, self.hdr, self.version, self.endian)
        self.map.read_data()

    def _version(self, filelike):
        """
        Get the rdbm.h revision number from first 4 bytes.

        Check against known revisions to confirm validity, and to infer
        endian-ness of the source data.

        Previously, this was performed in _major_version, which also maps to
        major platform number -- but this overlooks relevant differences across
        certain minor revisions

        """

        rev_little = RevisionNumLittle()
        rev_big    = RevisionNumBig()
        filelike.seek(0)
        filelike.readinto(rev_little)
        filelike.seek(0)
        filelike.readinto(rev_big)

        rev_little = rev_little.revision
        rev_big    = rev_big.revision

        known_revisions = \
            [
                7, 8, 9, 10, 11,
                14.0, 14.1, 14.2, 14.3,
                15.000, 15.001, 16.000,
                20.001, 20.002, 20.003, 20.004, 20.005, 20.006, 20.007,
                24.000,
                25.001, 25.002, 25.003, 25.004,
                26.000, 26.001, 26.002,
                27.000, 27.001,
                28.000, 28.002, 28.003,
                29.1,
                30.0
            ]

        # Note that caution is needed for float comparisons, given the
        # possibility of rounding errors. Although python usually handles this
        # reasonably, we use np.isclose here to be on the safe side.

        # Check first against little-endian interpretation (most likely for
        # recent implementations), then re-check against big-endian for older
        # configurations

        if any(np.isclose(rev_little, known_revisions)):
            # little-endian: most reasonably recent implementations
            self.endian = 'little'
            version = rev_little
        elif any(np.isclose(rev_big, known_revisions)) and rev_big < 11:
            # certain earlier revisions on SGI, big-endian
            self.endian = 'big'
            version = rev_big
        else:
            raise UnknownPfile("Unknown header structure for revision %s" % rev_little)

        return version

    def _major_version(self, filelike):
        """
        Get the rdbm.h revision number from first 4 bytes. Then map the rdbm
        revision to a platform number (e.g. 11.x, 12.x, etc.)

        """

        version = self._version(filelike)

        return int(np.trunc(version))

    def dump_header_strarr(self):

        dumped = self._dump_struct(self.hdr)
        strarr = []
        for info in dumped:
            if (info.label.find("pad") == 0):
                continue
            # needed this because Gregor's UID values had some odd chars which
            # caused errors on the import of Probep data
            try:
                val = info.value
                # needed this because Pom's UID values were causing errors
                # when VIFF was read back in
                if info.label == 'rhe_study_uid':
                    val = ' '
                if info.label == 'rhs_series_uid':
                    val = ' '
                if info.label == 'rhs_landmark_uid':
                    val = ' '
                if info.label == 'rhi_image_uid':
                    val = ' '
                val = str(val)
            except Exception:
                val = ' '
            strarr.append(str(info.label) + '        ' + val)

        return strarr

    def dump_header(self):

        dumped = self._dump_struct(self.hdr)
        writer = csv.writer(sys.stdout, delimiter="\t")
        writer.writerow(["\n\nHeader All  -  field", "value"])
        for info in dumped:
            if (info.label.find("pad") == 0):
                continue
            writer.writerow([info.label, str(info.value)])

    def _dump_struct(self, struct, include_structs=False):
        """
        Recursively travels through a ctypes.Structure and returns a list of
        namedtuples, containing label, depth, value, size, and offset.
        If include_structs is true, output will include lines for individual
        structures and their sizes and offsets -- not just non-structure fields.
        """
        output = []
        self._dump_struct_rec(struct, output, include_structs)
        return output

    def _dump_struct_rec(self, struct, output, include_structs=False, prefix='', depth=0, base_offset=0):
        """
        Internal recursive method for dumping structures.
        Appends to the "output" parameter.

        """
        struct_class = type(struct)
        if include_structs:
            output.append(StructInfo(
                f"{prefix} ({struct_class.__name__})",
                depth, '', str(struct_class), ctypes.sizeof(struct_class), base_offset))
        for f in struct._fields_:
            name = f[0]
            field_type = f[1]
            field_meta = getattr(struct_class, name)
            field = getattr(struct, name)
            cur_prefix = f"{prefix}{name}."
            field_offset = base_offset + field_meta.offset
            if isinstance(field, ctypes.Structure):
                self._dump_struct_rec(field, output, include_structs, cur_prefix, depth + 1, field_offset)
            else:
                label = prefix + name
                output.append(StructInfo(label, depth, field, field_type.__name__, field_meta.size, field_offset))


# ------------------------------------------------------------------------------
# originally ge_pfile_mapper.py
# ------------------------------------------------------------------------------


class PfileMapper:

    def __init__(self, file_name, hdr, version, endian):
        """
        Given a file name, its header, version number and endianness, this
        class will parse the data section of the file for the suppressed and
        unsuppressed data.

        All 'timePts' (aka. FID data arrays) are stored in the raw_data
        attribute. It is a numpy ndarray with shape of:

        [cols, rows, slices, numTimePts, numCoils, numSpecPts], np.complex64

        For SVS data, cols, rows and slices are all equal to 1.

            - raw_suppressed   is a view onto the  water suppressed fids data
            - raw_unsuppressed is a view onto the water unsuppressed fids data
            - avg_suppressed and avg_unsuppressed are numpy arrays where the
                  relevant raw_ views have been summed along the numTimePts
                  dimension. shape = [cols, rows, slices, numCoils, numSpecPts]

        For non-SVS data, only the raw_data attribute has data in it.

        History:

        Derived from SIVIC file svkGEPFileMapper.cc which was used to map data
        from PROBE-P and PRESSCSI P-files.  SIVIC has other mapper classes for
        other types of P-file data. I will plan on using this model here, too.

        """

        self.file_name = file_name
        self.hdr       = hdr
        self.version   = version
        self.endian    = endian
        self.is_svs    = False

        self.raw_data         = None
        self.raw_suppressed   = None
        self.avg_suppressed   = None
        self.raw_unsuppressed = None
        self.avg_unsuppressed = None

    @property
    def get_select_box_center(self):
        """
        Center position is taken from user variables.  The Z "slice"
        position used to be taken from the image header "image.loc",
        but with the LX architecture, this held the table position only,
        so if Graphic RX was used to introduce an offset, it wouldn't
        be successfully extracted.

        """
        center0 = self.hdr.rhi_user11
        center1 = self.hdr.rhi_user12
        center2 = self.hdr.rhi_user13

        return np.array([center0, center1, center2])

    @property
    def get_select_box_size(self):

        boxsize = np.array([0.0, 0.0, 0.0])
        dcos = self.get_dcos

        if self.version > 9:

            lMax   = 0
            pMax   = 0
            sMax   = 0
            lIndex = 0
            pIndex = 0
            sIndex = 0
            for i in range(3):
                if abs(dcos[i][0]) > lMax:
                    lIndex = i
                    lMax = abs(dcos[i][0])
                if abs(dcos[i][1]) > pMax:
                    pIndex = i
                    pMax = abs(dcos[i][1])
                if abs(dcos[i][2]) > sMax:
                    sIndex = i
                    sMax = abs(dcos[i][2])

            boxsize[lIndex] = self.hdr.rhi_user8
            boxsize[pIndex] = self.hdr.rhi_user9
            boxsize[sIndex] = self.hdr.rhi_user10

        else:

            boxsize[0] = self.hdr.rhr_roilenx
            boxsize[1] = self.hdr.rhr_roileny
            boxsize[2] = self.hdr.rhr_roilenz

            if self.is_swap_on:
                ftemp = boxsize[0]
                boxsize[0] = boxsize[1]
                boxsize[1] = ftemp

        return boxsize

    @property
    def get_voxel_spacing(self):
        """
        Get the voxel spacing in 3D. Note that the slice spacing may include
        a skip.
        Swaps the FOV if necessary based on freq_dir setting.

        """
        user19 = self.hdr.rhi_user19
        voxspace = np.array([0.0, 0.0, 0.0])

        if (user19 > 0) and (self.version > 9):
            voxspace[0] = user19
            voxspace[1] = user19
            voxspace[2] = user19
        else:
            fov  = self.get_fov
            nvox = self.get_num_voxels
            voxspace[0] = fov[0] / nvox[0]
            voxspace[1] = fov[1] / nvox[1]
            voxspace[2] = fov[2] / nvox[2]

        return voxspace

    @property
    def get_fov(self):
        fov  = np.array([0.0, 0.0, 0.0])
        nvox = self.get_num_voxels

        dfov = self.hdr.rhi_dfov

        if self.version > 9:

            fov[0] = dfov
            fov[1] = dfov

            # 2D case vs 3D cases
            if self.is_2d:
                fov[2] = self.hdr.rhi_user10
            else:
                fov[2] = self.hdr.rhi_scanspacing * self.hdr.rhr_zcsi
        else:
            fov[0] = self.hdr.rhr_rh_user7
            fov[1] = self.hdr.rhr_rh_user8
            fov[2] = self.hdr.rhr_rh_user9

        #  Anisotropic voxels:
        if (self.version > 9) and (nvox[0] != nvox[1]):

            # CSI has already been reordered if needed - so fov  calculated
            # with this CSI will not need reordering, need next power of 2:
            xdim = int(pow(2, math.ceil(math.log(nvox[0], 2))))
            ydim = int(pow(2, math.ceil(math.log(nvox[1], 2))))

            if (ydim > xdim):
                fov_spatial_resolution = dfov / ydim
            else:
                fov_spatial_resolution = dfov / xdim

            fov[1] = fov_spatial_resolution * ydim
            fov[0] = fov_spatial_resolution * xdim

        elif self.is_swap_on:

            #  Swap the FOV if necessary based on freq dir:
            temp = fov[0]
            fov[0] = fov[1]
            fov[1] = temp

        return fov

    @property
    def get_num_voxels(self):
        """
        Get the 3D spatial dimensionality of the data set
        Returns an int array with 3 dimensions.  Swaps
        if necessary based on freq_dir setting.

        """
        nvox = np.array([0, 0, 0])

        if self.hdr.rhr_rh_file_contents == 0:
            nvox[0] = 1
            nvox[1] = 1
            nvox[2] = 1
        else:
            nvox[0] = int(self.hdr.rhr_xcsi)
            nvox[1] = int(self.hdr.rhr_ycsi)
            nvox[2] = int(self.hdr.rhr_zcsi)

        #  Swap dimensions if necessary:
        if self.is_swap_on:
            temp = nvox[0]
            nvox[0] = nvox[1]
            nvox[1] = temp

        return nvox

    @property
    def get_dcos(self):
        dcos = np.zeros([3, 3], float)

        dcos[0][0] = (self.hdr.rhi_trhc_R - self.hdr.rhi_tlhc_R)
        dcos[0][1] = (self.hdr.rhi_trhc_A - self.hdr.rhi_tlhc_A)
        dcos[0][2] =  (self.hdr.rhi_trhc_S - self.hdr.rhi_tlhc_S)

        dcosLengthX = np.sqrt(dcos[0][0] * dcos[0][0]
                              + dcos[0][1] * dcos[0][1]
                              + dcos[0][2] * dcos[0][2])

        dcos[0][0] /= dcosLengthX
        dcos[0][1] /= dcosLengthX
        dcos[0][2] /= dcosLengthX

        dcos[1][0] = (self.hdr.rhi_brhc_R - self.hdr.rhi_trhc_R)
        dcos[1][1] = (self.hdr.rhi_brhc_A - self.hdr.rhi_trhc_A)
        dcos[1][2] =  (self.hdr.rhi_brhc_S - self.hdr.rhi_trhc_S)

        dcosLengthY = np.sqrt(dcos[1][0] * dcos[1][0]
                              + dcos[1][1] * dcos[1][1]
                              + dcos[1][2] * dcos[1][2])

        dcos[1][0] /= dcosLengthY
        dcos[1][1] /= dcosLengthY
        dcos[1][2] /= dcosLengthY

        # third row is the vector product of the first two rows
        # actually, -1* vector product, at least for the axial and axial oblique
        # which is all that we support now
        dcos[2][0] = - dcos[0][1] * dcos[1][2] + dcos[0][2] * dcos[1][1]
        dcos[2][1] = - dcos[0][2] * dcos[1][0] + dcos[0][0] * dcos[1][2]
        dcos[2][2] = - dcos[0][0] * dcos[1][1] + dcos[0][1] * dcos[1][0]

        return dcos

    @property
    def is_swap_on(self):
        """ Is frequency direction swapped? """
        if self.hdr.rhi_freq_dir != 1:
            return True
        else:
            return False

    @property
    def is_2d(self):
        """ Is this a 2D or 3D data set (spatial dimensions)? """
        is2D = False
        ndims = self.hdr.rhr_csi_dims

        if ndims == 0:
            if self.hdr.rhr_xcsi >= 0:
                ndims += 1
            if self.hdr.rhr_ycsi >= 0:
                ndims += 1
            if self.hdr.rhr_zcsi >= 0:
                ndims += 1

        if ndims == 2:
            is2D = True

        return is2D

    @property
    def is_chop_on(self):
        """ Is data chopped? """
        chop  = False
        nex   = self.hdr.rhi_nex
        necho = self.hdr.rhi_numecho

        if (math.ceil(nex) * necho) <= 1:
            chop = True

        return chop

    @property
    def get_frequency_offset(self):
        """ Returns the spectral frequency offset """
        if self.version > 9:
            return 0.0
        else:
            return self.hdr.rhr_rh_user13

    @property
    def get_center_from_raw_file(self):
        """
        Gets the center of the acquisition grid.  May vary between sequences.

        """
        center = np.array([0.0, 0.0, 0.0])
        if self.version < 11:
            center[0] = 0
            center[1] = 0
            center[2] = self.hdr.rhi_user13
        else:
            center[0] = -1 * self.hdr.rhi_user11
            center[1] = -1 * self.hdr.rhi_user12
            center[2] = self.hdr.rhi_user13

        return center

    @property
    def get_num_coils(self):
        """ Determine number of coils of data in the PFile. """
        ncoils = 0
        for i in range(4):
            start_rcv = getattr(self.hdr, "rhr_rh_dab[" + str(i) + "]_start_rcv")
            stop_rcv  = getattr(self.hdr, "rhr_rh_dab[" + str(i) + "]_stop_rcv")

            if (start_rcv != 0) or (stop_rcv != 0):
                ncoils += (stop_rcv - start_rcv) + 1

        #  Otherwise 1
        if ncoils == 0:
            ncoils = 1

        return int(ncoils)

    @property
    def get_num_time_points(self):
        """
        Determine number of time points in the PFile.
        Number of time points is determined from the file size,
        number of voxels and number of coils.
        """
        passSize      = float(self.hdr.rhr_rh_raw_pass_size)
        numCoils      = float(self.get_num_coils)
        numVoxels     = float(self.get_num_voxels_in_vol)  # noqa: F841
        dataWordSize  = float(self.hdr.rhr_rh_point_size)
        numFreqPoints = float(self.hdr.rhr_rh_frame_size)
        kSpacePoints  = float(self.get_num_kspace_points)

        numTimePoints = int((passSize) / (numCoils * 2 * dataWordSize * numFreqPoints) - 1) / kSpacePoints

        # bjs - added this after Pom's fidcsi 13C data came up with 0 here
        if numTimePoints <= 0:
            numTimePoints = 1

        return int(numTimePoints)

    @property
    def get_num_dummy_scans(self):
        """
        Determine number of dummy scans (FIDs) in the data block.
        This is the difference between the raw pass size and the
        expected size of the data based on numCoils, numTimePts, numKSpacePts
        and numFreqPts.

        """
        passSize         = self.hdr.rhr_rh_raw_pass_size
        numCoils         = self.get_num_coils
        numTimePoints    = self.get_num_time_points
        numSampledVoxels = self.get_num_kspace_points
        numFreqPoints    = self.hdr.rhr_rh_frame_size
        dataWordSize     = self.hdr.rhr_rh_point_size

        dataRepresentation = "COMPLEX"  # this was hard set in DcmHeader code
        if (dataRepresentation == "COMPLEX"):
            numComponents = 2
        else:
            numComponents = 1

        #  Calc the diff between the size of the data buffer and the number of real data points
        #  then divide by the number of bytes in a single fid to get the number of dummy FIDs
        numDummyScans = passSize - (numCoils * numTimePoints * numSampledVoxels
                                    * numFreqPoints * numComponents * dataWordSize)

        numDummyScans = numDummyScans / (numFreqPoints * numComponents * dataWordSize)

        return int(numDummyScans)

    @property
    def get_num_frames(self):
        """ Number of frames is number of slices * numCoils * numTimePoints """

        nvox = self.get_num_voxels
        nframes = nvox[2] * self.get_num_coils * self.get_num_time_points

        return int(nframes)

    @property
    def get_num_voxels_in_vol(self):

        nvox = self.get_num_voxels

        return int(nvox[0] * nvox[1] * nvox[2])

    @property
    def get_num_kspace_points(self):
        """
        Determine the number of sampled k-space points in the data set.
        This may differ from the number of voxels in the rectilinear grid,
        for example if elliptical or another non-rectangular acquisition
        sampling strategy was employed.  GE product sequences pad the
        reduced k-space data with zeros so the number of k-space points
        is the same as the number of voxels, but that may not be true for
        custom sequences.

        """
        return int(self.get_num_voxels_in_vol)

    @property
    def was_index_sampled(self):
        """
        Determines whether a voxel (index) was sampled (or a zero padded
        point is present in the data set), or not, i.e. was it within
        the elliptical sampling volume if reduced k-space elliptical sampling
        was used. Could be extended to support other sparse sampling
        trajectories. Note that for product sequences this always returns true
        since GE  zero-pads reduced k-space data to a full rectilinear grid.

        """
        return True

    @property
    def get_number_unsuppressed_acquisitions(self):
        """
        For single voxel acquisitions, return the number of
        unsuppressed acquisitions.

        """
        nex = self.hdr.rhi_nex
        return int(16 / nex)

    @property
    def get_number_suppressed_acquisitions(self):
        """
        For single voxel acquisitions, return the number of
        suppressed acquisitions.

        """
        nex   = self.hdr.rhi_nex
        user4 = self.hdr.rhi_user4
        return int(user4 / nex)

    def add_dummy(self, offset, coilNum, timePt):
        """
        Determine whether to add a dummy scan. The assumption is that
        the number of dummy scans should be equal to the number of coils
        or numCoils * numTimePts (e.g. for a spectral editing sequence).
        If true, then the an FID worth of data should be skipped over when
        reading data (e.g. frame_size * numComponents, or numFreqPts * numComponents)
        """
        numDummyScans    = self.get_num_dummy_scans
        numCoils         = self.get_num_coils
        numTimePoints    = self.get_num_time_points
        numSampledVoxels = self.get_num_kspace_points
        numFreqPoints    = self.hdr.rhr_rh_frame_size
        numComponents    = 2
        numPointsPerFID  = numFreqPoints * numComponents

        #  subtract the number of dummy words from the current offset to see if another
        #  dummy scan should be skipped or not

        if numDummyScans == numCoils:
            numWordsBetweenDummies = numSampledVoxels * numPointsPerFID * numTimePoints
            offset = offset - (coilNum * numPointsPerFID)
            # additional time points have an offset that includes the per-coil dummy
            if timePt > 1:
                offset = offset - numPointsPerFID

        elif (numDummyScans == (numCoils * numTimePoints)):
            numWordsBetweenDummies = numSampledVoxels * numPointsPerFID
            offset = offset - (coilNum * numPointsPerFID) - ((coilNum + timePt) * numPointsPerFID)

        elif numDummyScans == 0:    # bjs - added for fidcsi 13C data from Pom
            return False

        else:
            numWordsBetweenDummies = None
            # "ERROR: Can not determine placement of dummy scans in raw file reader. \n"

        addDummy = False
        if ((offset % numWordsBetweenDummies) == 0):
            addDummy = True

        return addDummy

    def get_xyz_indices(self, dataIndex):
        """
        If swapping is turned on, the data will need to get mapped correctly
        from the input data buffer read from disk (specData) to the correct
        svkImageData arrays. If swap is true, then the data indices are swapped
        and ky is flipped.

        """

        numVoxels = self.get_num_voxels

        z = int(dataIndex / (numVoxels[0] * numVoxels[1]))

        if self.is_swap_on:

            # If swap is on use numVoxels[1] for x dimension and numVoxels[0] for y dimension
            x = int((dataIndex - (z * numVoxels[0] * numVoxels[1])) / numVoxels[1])

            # In addition to swapping reverse the y direction
            y = numVoxels[1] - int(dataIndex % numVoxels[1]) - 1

        else:
            x = int(dataIndex % numVoxels[0])
            y = int((dataIndex - (z * numVoxels[0] * numVoxels[1])) / numVoxels[0])

        return x, y, z

    @staticmethod
    def get_center_from_origin(origin, numVoxels, voxelSpacing, dcos):
        """
        Calculates the LPS center from the origin(toplc).

        """
        center = np.array([0.0, 0.0, 0.0])

        for i in range(3):
            center[i] = origin[i]
            for j in range(3):
                center[i] += dcos[j][i] * voxelSpacing[j] * (numVoxels[j] / 2.0 - 0.5)
        return center

    @staticmethod
    def get_origin_from_center(center, numVoxels, voxelSpacing, dcos):
        """
        Calculates the LPS origin (toplc) from the center.

        """
        origin = np.array([0.0, 0.0, 0.0])

        for i in range(3):
            origin[i] = center[i]
            for j in range(3):
                origin[i] -= dcos[j][i] * voxelSpacing[j] * (numVoxels[j] / 2.0 - 0.5)
        return origin

    def read_data(self):
        """
        This method reads data from the pfile and puts the data into
        the CellData arrays. If elliptical k-space sampling was used,
        the data is zero-padded.  Other reduced k-space sampling
        strategies aren't supported yet.

        """

        numCoils        = self.get_num_coils
        numTimePts      = self.get_num_time_points
        numSpecPts      = self.hdr.rhr_rh_frame_size
        numFreqPts      = numSpecPts
        numComponents   =  2
        dataWordSize    = self.hdr.rhr_rh_point_size

        numBytesInVol   = self.get_num_kspace_points * numSpecPts * numComponents * dataWordSize
        numBytesPerCoil = numBytesInVol * numTimePts

        numPtsPerSpectrum = numSpecPts * numComponents

        #  one dummy spectrum per volume/coil:
        numDummyBytes = self.get_num_dummy_scans * numPtsPerSpectrum * dataWordSize
        numDummyBytesPerCoil = int(numDummyBytes / numCoils)

        numBytesPerCoil += numDummyBytesPerCoil

        #  Only read in one coil at a time to reduce memory footprint
        specData_size = int(np.round((numBytesPerCoil / dataWordSize), 0))
        specData = np.zeros([specData_size, ], float)

        try:
            readOffset = self.hdr.rhr_rdb_hdr_off_data
            filelike = open(self.file_name, 'rb')
            filelike.seek(0)
            filelike.seek(readOffset)
        except Exception:
            pass
            # "ERROR: Exception opening/reading file " << pFileNames->GetValue(0) << " => " << e.what() << endl;

        numVoxels       = self.get_num_voxels
        cols            = numVoxels[0]
        rows            = numVoxels[1]
        slices          = numVoxels[2]
        arraysPerVolume = cols * rows * slices

        #  Preallocate data arrays. The API only permits dynamic assignment at end of CellData, so for
        #  swapped cases where we need to insert data out of order they need to be preallocated.

        data = np.zeros([cols, rows, slices, numTimePts, numCoils, numSpecPts], np.complex64)

        #  Blank scan prepended to data blocks.
        dummyOffset = numPtsPerSpectrum

        #  If Chop On, then reinitialize chopVal:
        chopVal = 1
        if self.is_chop_on:
            chopVal = -1

        #  pFileOOffset is the offset of the current data set within the
        #  pFile (i.e. a global data offset).  #  a given coil.
        pFileOffset = 0

        for coilNum in range(numCoils):
            #  Offset is the offset within the current block of loaded data (ie. within a given coil)
            offset = 0

            tempData = None
            if dataWordSize == 4:
                tempData = np.fromfile(filelike, dtype='i4', count=int(numBytesPerCoil / dataWordSize))
            elif dataWordSize == 2:
                tempData = np.fromfile(filelike, dtype='i2', count=int(numBytesPerCoil / dataWordSize))
            if self.endian != 'little':
                tempData.byteswap(True)     # swap in-place
            specData = tempData.astype(np.float32)

            for timePt in range(numTimePts):

                #  Should a dummy scan be skipped over?
                if self.add_dummy(pFileOffset, coilNum, timePt):
                    offset      += dummyOffset
                    pFileOffset += dummyOffset

                for arrayNumber in range(arraysPerVolume):
                    x, y, z = self.get_xyz_indices(arrayNumber)

                    #  if k-space sampling was used check if the point was sampled, or if it needs
                    #  to be zero-padded in the grid.
                    #  if zero-padding don't increment the data pointer offset.

                    wasSampled = self.was_index_sampled

                    #  Default, chop is off, so multiply all values by 1
                    #  Only chop sampled data points.
                    if self.is_chop_on and wasSampled:
                        chopVal *= -1

                    if wasSampled:
                        tempFloat = specData[offset:offset + numFreqPts * numComponents]
                        data[x, y, z, timePt, coilNum, :] = chopVal * tempFloat.view(np.complex64)
                    else:
                        pass  # do nothing because array is initialized to zeros

                    if wasSampled:
                        offset      += numPtsPerSpectrum
                        pFileOffset += numPtsPerSpectrum

        filelike.close()

        self.raw_data = data

        #  Modify the data loading behavior.  For single voxel multi-acq data
        #  this means return the averaged (suppressed data, if applicable).

        numUnsuppressed = self.get_number_unsuppressed_acquisitions
        numSuppressed   = self.get_number_suppressed_acquisitions

        # bjs - changed numTimePoints to >= 1 below due to Pom's fidcsi 13C data
        # bjs - changed numSuppressed to >= 1 below due to oslaser dv26 data

        # Check if single voxel
        self.is_svs = False
        if ((numVoxels[0] * numVoxels[1] * numVoxels[2] == 1) and (numTimePts >= 1) and (numSuppressed >= 1)):
            self.is_svs = True

        if self.is_svs:
            self.raw_unsuppressed = self.raw_data[:, :, :, 0:numUnsuppressed, :, :]
            self.raw_suppressed   = self.raw_data[:, :, :, numUnsuppressed:, :, :]
            self.avg_unsuppressed = np.sum(self.raw_unsuppressed, axis=3) / float(numUnsuppressed)
            self.avg_suppressed   = np.sum(self.raw_suppressed, axis=3) / float(numSuppressed)
            self.phase_first_point_deg = np.angle(self.avg_unsuppressed[0, 0, 0, :, 0], True)
        else:
            self.raw_suppressed   = None
            self.avg_suppressed   = None
            self.raw_unsuppressed = None
            self.avg_unsuppressed = None
            self.phase_of_first_point_deg = None


class PfileMapperProbeSL(PfileMapper):

    def __init__(self, file_name, hdr, version, endian):
        """
        Quick attempt to make sense of the probe-sl CSI data

        """
        PfileMapper.__init__(self, file_name, hdr, version, endian)

    @property
    def get_voxel_spacing(self):
        """
        Get the voxel spacing in 3D. Note that the slice spacing may include
        a skip.
        Swaps the FOV if necessary based on freq_dir setting.

        """
        voxspace = np.array([0.0, 0.0, 0.0])

        fov  = self.get_fov
        nvox = self.get_num_voxels
        voxspace[0] = fov[0] / nvox[0]
        voxspace[1] = fov[1] / nvox[1]
        voxspace[2] = fov[2] / nvox[2]

        return voxspace


class PfileMapperSlaser(PfileMapper):

    def __init__(self, file_name, hdr, version, endian):
        """
        This info was provided by Ralph Noeske at GE who wrote the sLASER
        sequence in collaboration with Gulin Oz. Needed to get a few
        parameters from different header locations, but otherwise the basic
        code seems to work OK.

        rdb_hdr_user0 = rhuser0 = 6024 (sampling frequency)
        rdb_hdr_user4 = rhuser4 = 64 (number of averages)
        rdb_hdr_user19 = rhuser19 = 8 (number of reference frames)

        image_user8 = opuser8 is used as starting for voxel dimension and location (opuser8,9,10,11,12,13)

        The sub-echo-times are TE1 = 8ms and TE2 = 12ms (opuser20 and 21).


        """
        PfileMapper.__init__(self, file_name, hdr, version, endian)

    @property
    def get_number_unsuppressed_acquisitions(self):
        """
        For single voxel acquisitions, return the number of
        unsuppressed acquisitions.

        """
        user19 = self.hdr.rhr_rh_user19
        return int(user19)

    @property
    def get_number_suppressed_acquisitions(self):
        """
        For single voxel acquisitions, return the number of
        suppressed acquisitions.

        """
        user4 = self.hdr.rhr_rh_user4
        return int(user4)


class PfileMapperGaba(PfileMapper):

    def __init__(self, file_name, hdr, version, endian):
        """
        MEGA-PRESS sequence.
        WTC first saw data from Nottingham but it seems general and appears in
        e.g. the Big GABA dataset.

        Without some more knowledgeable input I can't currently square the logic
        of BJS's mapper classes with the logic from Gannet/Osprey etc.
        Therefore, this is really a hack that uses the PfileMapper orientation logic
        but otherwise uses the logic from Gannet/Osprey.

        Thus, most work is done in the overloaded read_data method and functions like
        get_num_voxels_in_vol are currently meaningless. I would like to square both
        methods in the future.
        """
        PfileMapper.__init__(self, file_name, hdr, version, endian)

    def read_data(self):
        """Function that contains all the data loading logic for the 'gaba'
        sequence.

        This is currently a reimplementation of the Gannet/Osprey GELoad.m
        function. Therefore, the logic is unlike the other mappers.
        The suppressed and unsuppressed data can be fetched from the raw_suppressed
        and raw_unsuppressed property.
        """

        nechoes = self.hdr.rhr_rh_nechoes
        nex = self.hdr.rhr_rh_navs
        nframes = self.hdr.rhr_rh_nframes
        npoints = self.hdr.rhr_rh_da_xres
        nrows = self.hdr.rhr_rh_da_yres

        dataframes = self.hdr.rhr_rh_user4 / nex
        refframes = int(self.hdr.rhr_rh_user19)

        nreceivers = self.get_num_coils
        dataWordSize = self.hdr.rhr_rh_point_size

        # MM: CV24 is a control variable (CV) that defines several advanced options not stored in the other CVs
        cv24 = self.hdr.rhi_user24

        # Check if single voxel
        numVoxels = self.get_num_voxels
        self.is_svs = False
        if numVoxels[0] * numVoxels[1] * numVoxels[2] == 1:
            self.is_svs = True

        # Data loading
        # Compute size (in bytes) of data
        data_elements = npoints * 2
        totalframes = nrows * nechoes
        data_elements = data_elements * totalframes * nreceivers

        readOffset = self.hdr.rhr_rdb_hdr_off_data
        with open(self.file_name, 'rb') as filelike:
            filelike.seek(0)
            filelike.seek(readOffset)

            if dataWordSize == 4:
                tempData = np.fromfile(filelike, dtype='i4', count=int(data_elements))
            elif dataWordSize == 2:
                tempData = np.fromfile(filelike, dtype='i2', count=int(data_elements))

        if self.endian != 'little':
            tempData.byteswap(True)     # swap in-place

        tempData = tempData.astype(np.float32)
        tempData = tempData.view(np.complex64)

        if nechoes == 1:
            if int(dataframes + refframes) != nframes:
                mult = 1
                dataframes *= nex
                refframes = int(nframes - dataframes)
            else:
                mult = 1 / nex

            tempData = tempData.reshape((1, 1, 1, nreceivers, totalframes, npoints))
            self.raw_data = np.swapaxes(tempData, -1, -3)
            self.raw_unsuppressed = self.raw_data[:, :, :, :, 1:(refframes + 1), :] * mult
            self.raw_suppressed = self.raw_data[:, :, :, :, (refframes + 1):, :] * mult / 2

        else:
            if int(dataframes + refframes) != nframes:
                mult = nex / 2
                multw = nex
                noadd = 1
                dataframes *= nex
                refframes = nframes - dataframes
            else:
                # MM: Change mult to match latest code in Gannet (2020)
                # Previous factors:
                #   mult: 1 (RTN 2017), nex / 2 (RTN 2016)
                #   multw: 1 / nex (RTN 2017), 1 (RTN 2016)
                mult = 1 / nex
                multw = 1
                noadd = 0

            if totalframes != (dataframes + refframes + 1) * nechoes:
                raise ValueError('# of totalframes not same as (dataframes + refframes + 1) * nechoes')

            tempData = tempData.reshape((1, 1, 1, nreceivers, totalframes, npoints))
            self.raw_data = np.swapaxes(tempData, -1, -3)

            # Marked as MM (180404) in GELoad.m
            X1, X2 = np.meshgrid(np.arange(refframes), np.arange(nechoes))
            X1 = X1.T.ravel()
            X2 = X2.T.ravel()
            # Do not apply any phase cycling correction if the receiver phase toggle in sLASER was set
            if cv24 >= 16384:
                Y1 = np.ones_like(X1)
            else:
                Y1 = (-1) ** (noadd * X1)
            Y1 = np.moveaxis(np.broadcast_to(Y1, (1, 1, 1, npoints, nreceivers, Y1.size)), -1, -2)
            Y2 = (1 + (totalframes / nechoes) * X2 + X1).astype(int)

            self.raw_unsuppressed = self.raw_data[:, :, :, :, Y2, :] * Y1 * multw

            X1, X2 = np.meshgrid(np.arange(dataframes), np.arange(nechoes))
            X1 = X1.T.ravel()
            X2 = X2.T.ravel()
            # Do not apply any phase cycling correction if the receiver phase toggle in sLASER was set
            if cv24 >= 16384:
                Y1 = np.ones_like(X1)
            else:
                Y1 = (-1) ** (noadd * X1)
            Y1 = np.moveaxis(np.broadcast_to(Y1, (1, 1, 1, npoints, nreceivers, Y1.size)), -1, -2)
            Y2 = (1 + refframes + (totalframes / nechoes) * X2 + X1).astype(int)

            self.raw_suppressed = self.raw_data[:, :, :, :, Y2, :] * Y1 * mult

        # Test if this is the case of the jpress like sequence being used for a non-editing condition
        edit_waveform = self.hdr.rhi_user19
        if nechoes == 1 and edit_waveform == 0:
            # Editing conditions = 1 and there is no editing waveform
            self.raw_suppressed[:, :, :, :, 0::2, :] *= np.exp(1j * np.pi)
            self.raw_unsuppressed[:, :, :, :, 0::2, :] *= np.exp(1j * np.pi)

            # Rearrange axes to (x, y, z, t, coils, dynamics)
            self.raw_suppressed = np.moveaxis(self.raw_suppressed, (4, 5), (5, 4))
            self.raw_unsuppressed = np.moveaxis(self.raw_unsuppressed, (4, 5), (5, 4))
        else:
            # Editing condition, update nechoes if needed for special cases like probe-p-mega_rml
            if nechoes == 1:
                nechoes = 2

            # Up to this point we have simply replicated the logic of the GELoad function.
            # Now reorganise dimensions to give an editing dimension.
            # This means that this is done in a not particularly clear order, but it enables testing against
            # the matlab code.
            reorg_suppressed = []
            reorg_unsuppressed = []
            for ne in range(nechoes):
                reorg_suppressed.append(self.raw_suppressed[:, :, :, :, ne::nechoes, :])
                reorg_unsuppressed.append(self.raw_unsuppressed[:, :, :, :, ne::nechoes, :])

            reorg_suppressed = np.stack(reorg_suppressed, axis=-1)
            # Rearrange axes to (x, y, z, t, coils, dynamics, edit)
            self.raw_suppressed = np.moveaxis(reorg_suppressed, (4, 5), (5, 4))

            reorg_unsuppressed = np.stack(reorg_unsuppressed, axis=-1)
            # Rearrange axes to (x, y, z, t, coils, dynamics, edit)
            self.raw_unsuppressed = np.moveaxis(reorg_unsuppressed, (4, 5), (5, 4))



#!/usr/bin/env python

from setuptools import setup, find_packages
import versioneer
import yaml

with open("requirements.yml", "r") as stream:
    try:
        requirements = yaml.safe_load(stream)
        install_requires = requirements['dependencies']
    except yaml.YAMLError as exc:
        print(exc)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='spec2nii',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Multi-format in vivo MR spectroscopy conversion to NIFTI',
      author='Will Clarke',
      author_email='william.clarke@ndcn.ox.ac.uk',
      url='https://github.com/wtclarke/spec2nii',
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=find_packages(exclude=["tests*"]),
      install_requires=install_requires,
      classifiers=[
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11",
          "Programming Language :: Python :: 3.12",
          "License :: OSI Approved :: BSD License",
          "Operating System :: OS Independent"],
      python_requires='>=3.9',
      entry_points={"console_scripts": [
          "spec2nii = spec2nii.spec2nii:main"]},
      package_data={'spec2nii': ['bruker_properties.json',
                                 'bruker_fid_override.json'],
                    'spec2nii.GE': ['VESPA_LICENSE']}
      )



import numpy as np
from scipy.spatial.transform import Rotation


class NIFTIOrient:
    def __init__(self, affine):
        self.Q44 = affine
        qb, qc, qd, qx, qy, qz, dx, dy, dz, qfac = nifti_mat44_to_quatern(affine)
        self.qb = qb
        self.qc = qc
        self.qd = qd
        self.qx = qx
        self.qy = qy
        self.qz = qz
        self.dx = dx
        self.dy = dy
        self.dz = dz
        self.qfac = qfac


def calc_affine(angles, dimensions, shift):

    scalingMat = np.diag(dimensions)
    # Rotations appear to be intrinsic
    rot = Rotation.from_euler('XYZ', angles, degrees=True)
    m33 = rot.as_matrix() @ scalingMat
    m44 = np.zeros((4, 4))
    m44[0:3, 0:3] = m33
    m44[3, 3] = 1.0
    m44[0:3, 3] = shift

    return m44


def nifti_mat44_to_quatern(R):
    """4x4 affine to quaternion representation."""
    # offset outputs are read out of input matrix
    qx = R[0, 3]
    qy = R[1, 3]
    qz = R[2, 3]

    # load 3x3 matrix into local variables
    r11 = R[0, 0]
    r12 = R[0, 1]
    r13 = R[0, 2]
    r21 = R[1, 0]
    r22 = R[1, 1]
    r23 = R[1, 2]
    r31 = R[2, 0]
    r32 = R[2, 1]
    r33 = R[2, 2]

    # compute lengths of each column; these determine grid spacings
    xd = np.sqrt(r11 * r11 + r21 * r21 + r31 * r31)
    yd = np.sqrt(r12 * r12 + r22 * r22 + r32 * r32)
    zd = np.sqrt(r13 * r13 + r23 * r23 + r33 * r33)

    # if a column length is zero, patch the trouble
    if xd == 0.0:
        r11 = 1.0
        r21 = 0.0
        r31 = 0.0
        xd = 1.0
    if yd == 0.0:
        r22 = 1.0
        r12 = 0.0
        r32 = 0.0
        yd = 1.0
    if zd == 0.0:
        r33 = 1.0
        r13 = 0.0
        r23 = 0.0
        zd = 1.0

    # assign the output lengths
    dx = xd
    dy = yd
    dz = zd

    # normalize the columns
    r11 /= xd
    r21 /= xd
    r31 /= xd
    r12 /= yd
    r22 /= yd
    r32 /= yd
    r13 /= zd
    r23 /= zd
    r33 /= zd

    zd = r11 * r22 * r33\
        - r11 * r32 * r23\
        - r21 * r12 * r33\
        + r21 * r32 * r13\
        + r31 * r12 * r23\
        - r31 * r22 * r13
    # zd should be -1 or 1

    if zd > 0:  # proper
        qfac = 1.0
    else:  # improper ==> flip 3rd column
        qfac = -1.0
        r13 *= -1.0
        r23 *= -1.0
        r33 *= -1.0

    # now, compute quaternion parameters
    a = r11 + r22 + r33 + 1.0
    if a > 0.5:  # simplest case
        a = 0.5 * np.sqrt(a)
        b = 0.25 * (r32 - r23) / a
        c = 0.25 * (r13 - r31) / a
        d = 0.25 * (r21 - r12) / a
    else:  # trickier case
        xd = 1.0 + r11 - (r22 + r33)  # 4*b*b
        yd = 1.0 + r22 - (r11 + r33)  # 4*c*c
        zd = 1.0 + r33 - (r11 + r22)  # 4*d*d
        if xd > 1.0:
            b = 0.5 * np.sqrt(xd)
            c = 0.25 * (r12 + r21) / b
            d = 0.25 * (r13 + r31) / b
            a = 0.25 * (r32 - r23) / b
        elif yd > 1.0:
            c = 0.5 * np.sqrt(yd)
            b = 0.25 * (r12 + r21) / c
            d = 0.25 * (r23 + r32) / c
            a = 0.25 * (r13 - r31) / c
        else:
            d = 0.5 * np.sqrt(zd)
            b = 0.25 * (r13 + r31) / d
            c = 0.25 * (r23 + r32) / d
            a = 0.25 * (r21 - r12) / d

        if a < 0.0:
            b = -b
            c = -c
            d = -d

    qb = b
    qc = c
    qd = d
    return qb, qc, qd, qx, qy, qz, dx, dy, dz, qfac



'''Tests for Bruker format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs

# Data paths
bruker_path = Path(__file__).parent / 'spec2nii_test_data' / 'bruker'
data_path = bruker_path / '20201208_105201_lego_rod_1_3'


def test_fid(tmp_path):

    subprocess.check_call(['spec2nii', 'bruker',
                           '-f', 'fid',
                           '-m', 'FID',
                           '-d',
                           '-o', tmp_path,
                           '-j',
                           str(data_path)])

    # Img 5 - csi
    img = read_nifti_mrs(tmp_path / 'fid_FID_5.nii.gz')

    assert img.shape == (16, 16, 1, 1980)
    assert np.iscomplexobj(img.dataobj)
    assert 1 / img.header['pixdim'][4] == 4000.0

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert np.isclose(hdr_ext['SpectrometerFrequency'][0], 400.32251)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == str(data_path.absolute() / '5' / 'fid')
    assert 'method' in hdr_ext
    assert 'acqp' in hdr_ext

    # Img 6 - csi
    img = read_nifti_mrs(tmp_path / 'fid_FID_6.nii.gz')

    assert img.shape == (16, 16, 1, 1980)
    assert np.iscomplexobj(img.dataobj)
    assert 1 / img.header['pixdim'][4] == 4000.0

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert np.isclose(hdr_ext['SpectrometerFrequency'][0], 400.32251)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == str(data_path.absolute() / '6' / 'fid')
    assert 'method' in hdr_ext
    assert 'acqp' in hdr_ext

    # Img 9 - svs
    img = read_nifti_mrs(tmp_path / 'fid_FID_9.nii.gz')

    assert img.shape == (1, 1, 1, 1980)
    assert np.iscomplexobj(img.dataobj)
    assert np.isclose(1 / img.header['pixdim'][4], 4401.41)

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert np.isclose(hdr_ext['SpectrometerFrequency'][0], 400.32251)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == str(data_path.absolute() / '9' / 'fid')
    assert 'method' in hdr_ext
    assert 'acqp' in hdr_ext

    # Img 10 - svs
    img = read_nifti_mrs(tmp_path / 'fid_FID_10.nii.gz')

    assert img.shape == (1, 1, 1, 1980)
    assert np.iscomplexobj(img.dataobj)
    assert np.isclose(1 / img.header['pixdim'][4], 4401.41)

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert np.isclose(hdr_ext['SpectrometerFrequency'][0], 400.32251)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == str(data_path.absolute() / '10' / 'fid')
    assert 'method' in hdr_ext
    assert 'acqp' in hdr_ext


def test_2dseq(tmp_path):

    subprocess.check_call(['spec2nii', 'bruker',
                           '-f', '2dseq',
                           '-m', '2DSEQ',
                           '-d',
                           '-o', tmp_path,
                           '-j',
                           str(data_path)])

    # Img 5 - csi
    img = read_nifti_mrs(tmp_path / '2dseq_2DSEQ_5_2_lego_rod_3.nii.gz')

    assert img.shape == (16, 16, 1, 2048)
    assert np.iscomplexobj(img.dataobj)
    assert 1 / img.header['pixdim'][4] == 4000.0

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert np.isclose(hdr_ext['SpectrometerFrequency'][0], 400.32251)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == str(data_path.absolute() / '5' / 'pdata' / '2' / '2dseq')
    assert 'method' in hdr_ext
    assert 'visu_pars' in hdr_ext

    # Img 6 - csi
    img = read_nifti_mrs(tmp_path / '2dseq_2DSEQ_6_2_lego_rod_3.nii.gz')

    assert img.shape == (16, 16, 1, 2048)
    assert np.iscomplexobj(img.dataobj)
    assert 1 / img.header['pixdim'][4] == 4000.0

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert np.isclose(hdr_ext['SpectrometerFrequency'][0], 400.32251)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == str(data_path.absolute() / '6' / 'pdata' / '2' / '2dseq')
    assert 'method' in hdr_ext
    assert 'visu_pars' in hdr_ext

    # Img 9 - svs
    img = read_nifti_mrs(tmp_path / '2dseq_2DSEQ_9_2_lego_rod_3.nii.gz')

    assert img.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(img.dataobj)
    assert np.isclose(1 / img.header['pixdim'][4], 4401.41)

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert np.isclose(hdr_ext['SpectrometerFrequency'][0], 400.32251)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == str(data_path.absolute() / '9' / 'pdata' / '2' / '2dseq')
    assert 'method' in hdr_ext
    assert 'visu_pars' in hdr_ext

    # Img 10 - svs
    img = read_nifti_mrs(tmp_path / '2dseq_2DSEQ_10_2_lego_rod_3.nii.gz')

    assert img.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(img.dataobj)
    assert np.isclose(1 / img.header['pixdim'][4], 4401.41)

    hdr_ext_codes = img.header.extensions.get_codes()
    hdr_ext = json.loads(img.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert np.isclose(hdr_ext['SpectrometerFrequency'][0], 400.32251)
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == str(data_path.absolute() / '10' / 'pdata' / '2' / '2dseq')
    assert 'method' in hdr_ext
    assert 'visu_pars' in hdr_ext



'''Orientation tests for Philips SDAT/SPAR format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path

import pytest
from PIL import Image, ImageDraw

# Data paths
output_path = Path(__file__).parent / 'orientation_img'
output_path.mkdir(exist_ok=True)

philips_path = Path(__file__).parent / 'spec2nii_test_data' / 'philips'

svs_path_sdat = [philips_path / 'P1' / 'SV_PRESS_sh_6_2_raw_act.SDAT',
                 philips_path / 'P2' / 'SV_PRESS_sh_14_2_raw_act.SDAT',
                 philips_path / 'P3' / 'SV_PRESS_sh_4_2_raw_act.SDAT',
                 philips_path / 'P4' / 'SV_PRESS_sh_11_2_raw_act.SDAT']

svs_path_spar = [philips_path / 'P1' / 'SV_PRESS_sh_6_2_raw_act.SPAR',
                 philips_path / 'P2' / 'SV_PRESS_sh_14_2_raw_act.SPAR',
                 philips_path / 'P3' / 'SV_PRESS_sh_4_2_raw_act.SPAR',
                 philips_path / 'P4' / 'SV_PRESS_sh_11_2_raw_act.SPAR']

svs_path = [philips_path / 'P1',
            philips_path / 'P2',
            philips_path / 'P3',
            philips_path / 'P4']

t1_pos = [(103, 156, 167),
          (90, 126, 145),
          (112, 179, 150),
          (107, 52, 107)]


def get_concat_v(im1, im2):
    dst = Image.new('RGB', (max(im1.width, im2.width), im1.height + im2.height))
    dst.paste(im1, (0, 0))
    dst.paste(im2, (0, im1.height))
    return dst


def crop_and_flip_first_third_and_swap(img_in, swap='23'):
    width, height = img_in.size
    cropped_0 = img_in.crop((0, 180, int(width / 3) + 15, height - 180))
    cropped_0 = cropped_0.transpose(Image.FLIP_LEFT_RIGHT)
    cropped_1 = img_in.crop((int(width / 3) + 16, 180, 2 * int(width / 3) + 16, height - 180))
    cropped_2 = img_in.crop((2 * int(width / 3) + 17, 180, width, height - 180))

    out = Image.new('RGB', (width, height - 360))
    if swap == '23':
        out.paste(cropped_0, (0, 0))
        out.paste(cropped_2, (int(width / 3) + 16, 0))
        out.paste(cropped_1, (2 * int(width / 3) + 17, 0))
    elif swap == '123':
        out.paste(cropped_2, (0, 0))
        out.paste(cropped_0, (int(width / 3), 0))
        out.paste(cropped_1, (2 * int(width / 3) + 17, 0))
    return out


@pytest.mark.orientation
def test_svs_orientation(tmp_path):

    sub_images = []
    for idx, (sdat, spar, p_path, pos) in enumerate(zip(svs_path_sdat, svs_path_spar, svs_path, t1_pos)):
        subprocess.check_call(['spec2nii', 'philips',
                               '-f', 'svs',
                               '-o', tmp_path,
                               '-j',
                               str(sdat),
                               str(spar)])

        # Make fsleyes rendering
        subprocess.check_call(['fsleyes', 'render',
                               '-of', tmp_path / f'svs_{idx}.png',
                               '-vl', str(pos[0]), str(pos[1]), str(pos[2]),
                               '-xc', '0', '0', '-yc', '0', '0', '-zc', '0', '0',
                               '-hc', p_path / 'T1.nii.gz', '-dr', '-8200', '204600',
                               tmp_path / 'svs.nii.gz', '-ot', 'complex',
                               '-a', '50', '-cm', 'blue'])

        fsl_ss = Image.open(tmp_path / f'svs_{idx}.png')
        width, height = fsl_ss.size
        if idx == 2:
            swap = '123'
        else:
            swap = '23'
        fsl_ss_cropped = crop_and_flip_first_third_and_swap(fsl_ss, swap=swap)

        ss = Image.open(p_path / 'placement.png')
        width, height = ss.size
        ss_cropped = ss.crop((10, 110, width - 50, height / 2 - 30))

        fsl_ss_cropped = fsl_ss_cropped.resize(ss_cropped.size)
        sub_images.append(get_concat_v(fsl_ss_cropped, ss_cropped))

    final_img = Image.new('RGB', (sub_images[0].width * 2, sub_images[0].height * 2))
    draw = ImageDraw.Draw(final_img)
    for idx, si in enumerate(sub_images):
        c = idx % 2
        r = int(idx / 2)
        final_img.paste(si, (si.width * r, si.height * c))
        draw.text((10 + si.width * r, 10 + si.height * c), f'P{idx + 1}', (255, 0, 0))

    final_img.save(output_path / 'philips_svs.png')



'''Interpretation of the Siemens GSL functions.
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford '''

import numpy as np


def class_ori(sag_comp, cor_comp, tra_comp, debug):
    ''' Python implementation of IDEA-VB17/n4/pkg/MrServers/MrMeasSrv/SeqFW/libGSL/fGSLClassOri.cpp
    Function to determine whether a normal vector describes a sagittal, coronal or transverse slice.
    Result:
        CASE = 0: Sagittal
        CASE = 1: Coronal
        CASE = 2: Transverse

    :param  sag_comp:   Sagittal component of normal vector
    :param  cor_comp:   Coronal component of normal vector
    :param  tra_comp:   Transverse component of normal vector

    :return:    case (0=Sagittal, 1=Coronal or 2=Transverse)
    '''

    if debug:
        print(f'Normal vector = {sag_comp: 10.7f} {cor_comp: 10.7f} {tra_comp: 10.7f}.')

    # Compute some temporary values
    abs_sag_comp     = np.abs(sag_comp)
    abs_cor_comp     = np.abs(cor_comp)
    abs_tra_comp     = np.abs(tra_comp)

    eq_sag_cor = np.isclose(abs_sag_comp, abs_cor_comp)
    eq_sag_tra = np.isclose(abs_sag_comp, abs_tra_comp)
    eq_cor_tra = np.isclose(abs_cor_comp, abs_tra_comp)

    # Determine the slice orientation (sag, cor, tra)
    if ((eq_sag_cor              & eq_sag_tra)             |
            (eq_sag_cor              & (abs_sag_comp < abs_tra_comp)) |
            (eq_sag_tra              & (abs_sag_comp > abs_cor_comp)) |
            (eq_cor_tra              & (abs_cor_comp > abs_sag_comp)) |
            ((abs_sag_comp > abs_cor_comp)  & (abs_sag_comp < abs_tra_comp)) |
            ((abs_sag_comp < abs_cor_comp)  & (abs_cor_comp < abs_tra_comp)) |
            ((abs_sag_comp < abs_tra_comp)  & (abs_tra_comp > abs_cor_comp)) |
            ((abs_cor_comp < abs_tra_comp)  & (abs_tra_comp > abs_sag_comp))):

        if debug:
            print('Mainly transverse.')
        case = 2  # Transverse

    elif ((eq_sag_cor              & (abs_sag_comp > abs_tra_comp)) |
            (eq_sag_tra              & (abs_sag_comp < abs_cor_comp)) |
            ((abs_sag_comp < abs_cor_comp)  & (abs_cor_comp > abs_tra_comp)) |
            ((abs_sag_comp > abs_tra_comp)  & (abs_sag_comp < abs_cor_comp)) |
            ((abs_sag_comp < abs_tra_comp)  & (abs_tra_comp < abs_cor_comp))):

        if debug:
            print('Mainly coronal.')
        case = 1  # Coronal

    elif ((eq_cor_tra              & (abs_cor_comp < abs_sag_comp)) |
            ((abs_sag_comp > abs_cor_comp)  & (abs_sag_comp > abs_tra_comp)) |
            ((abs_cor_comp > abs_tra_comp)  & (abs_cor_comp < abs_sag_comp)) |
            ((abs_cor_comp < abs_tra_comp)  & (abs_tra_comp < abs_sag_comp))):

        if debug:
            print('Mainly sagittal.')
        case = 0  # Sagittal

    else:    # Invalid slice orientation...
        raise ValueError('Error: Invalid slice orientation')

    return case


def calc_prs(gs, phi, debug):
    ''' Python implementation of IDEA-VB17/n4/pkg/MrServers/MrMeasSrv/SeqFW/libGSL/fGSLCalcPRS.cpp
    Calculates the phase encoding and readout direction vectors

    :param gs: The GS vector (= slice normal vector)
    :param phi: The rotational angle around Gs

    :return: gp: phase direction vector
    :return: gr: read direction vector
    '''

    # PCS axes
    SAGITTAL   = 0
    CORONAL    = 1
    TRANSVERSE = 2

    # Start of function
    orientation = 0  # will be one of SAGITTAL, CORONAL or TRANSVERSE (0, 1, or 2)
    orientation = orientation + class_ori(gs[SAGITTAL], gs[CORONAL], gs[TRANSVERSE], debug)
    gp  = np.zeros((3), dtype=float)

    if orientation == TRANSVERSE:
        gp[0] = 0.0
        gp[1] = gs[2] * np.sqrt(1. / (gs[1] * gs[1] + gs[2] * gs[2]))
        gp[2] = -gs[1] * np.sqrt(1. / (gs[1] * gs[1] + gs[2] * gs[2]))
    elif orientation == CORONAL:
        gp[0] = gs[1] * np.sqrt(1. / (gs[0] * gs[0] + gs[1] * gs[1]))
        gp[1] = -gs[0] * np.sqrt(1. / (gs[0] * gs[0] + gs[1] * gs[1]))
        gp[2] = 0.0
    elif orientation == SAGITTAL:
        gp[0] = -gs[1] * np.sqrt(1. / (gs[0] * gs[0] + gs[1] * gs[1]))
        gp[1] = gs[0] * np.sqrt(1. / (gs[0] * gs[0] + gs[1] * gs[1]))
        gp[2] = 0.0
    else:
        raise ValueError('Invalid slice orientation returned from class_ori')

    # Calculate GR = GS x GP
    gr = np.zeros((3), dtype=float)
    gr[0] = gs[1] * gp[2] - gs[2] * gp[1]
    gr[1] = gs[2] * gp[0] - gs[0] * gp[2]
    gr[2] = gs[0] * gp[1] - gs[1] * gp[0]

    if debug:
        print('Before rotation around S:')
        print(f'GP = {gp[0]:10.7f} {gp[1]:10.7f} {gp[2]:10.7f}')
        print(f'GR = {gr[0]:10.7f} {gr[1]:10.7f} {gr[2]:10.7f}')
        print(f'GS = {gs[0]:10.7f} {gs[1]:10.7f} {gs[2]:10.7f}')

    # rotation
    if phi != 0.0:
        # Rotate by phi around the S axis
        if debug:
            tmp = phi * 180.0 / np.pi
            print(f'PHI = {tmp:10.7f}')

    gp[0] = np.cos(phi) * gp[0] - np.sin(phi) * gr[0]
    gp[1] = np.cos(phi) * gp[1] - np.sin(phi) * gr[1]
    gp[2] = np.cos(phi) * gp[2] - np.sin(phi) * gr[2]

    # Calculate new GR = GS x GP
    gr[0] = gs[1] * gp[2] - gs[2] * gp[1]
    gr[1] = gs[2] * gp[0] - gs[0] * gp[2]
    gr[2] = gs[0] * gp[1] - gs[1] * gp[0]

    if debug:
        print('After the Rotation around S:')
        print(f'GP = {gp[0]:10.7f} {gp[1]:10.7f} {gp[2]:10.7f}')
        print(f'GR = {gr[0]:10.7f} {gr[1]:10.7f} {gr[2]:10.7f}')
        print(f'GS = {gs[0]:10.7f} {gs[1]:10.7f} {gs[2]:10.7f}')

    return gp, gr



'''Tests for Philips SDAT/SPAR format conversion.

Copyright William Clarke, University of Oxford 2021
Subject to the BSD 3-Clause License.
'''

import subprocess
from pathlib import Path
import json

import numpy as np

from .io_for_tests import read_nifti_mrs
from nifti_mrs.nifti_mrs import NIFTI_MRS

# Data paths
philips_path = Path(__file__).parent / 'spec2nii_test_data' / 'philips'
svs_path_sdat = philips_path / 'P1' / 'SV_PRESS_sh_6_2_raw_act.SDAT'
svs_path_spar = philips_path / 'P1' / 'SV_PRESS_sh_6_2_raw_act.SPAR'

svs_ma_path_sdat = philips_path / 'spar_multi_avg' / 'multi_avg.sdat'
svs_ma_path_spar = philips_path / 'spar_multi_avg' / 'multi_avg.spar'

svs_edit_path_sdat = philips_path / 'HERCULES_spar_sdat' / 'HERCULES_Example_noID.sdat'
svs_edit_path_spar = philips_path / 'HERCULES_spar_sdat' / 'HERCULES_Example_noID.spar'

hyper_path_sdat = philips_path / 'hyper' / 'HBCD_HYPER_r5712_Export_WIP_HYPER_5_2_raw_act.SDAT'
hyper_path_spar = philips_path / 'hyper' / 'HBCD_HYPER_r5712_Export_WIP_HYPER_5_2_raw_act.SPAR'

hyper_ref_path_sdat = philips_path / 'hyper' / 'HBCD_HYPER_r5712_Export_WIP_HYPER_5_2_raw_ref.SDAT'
hyper_ref_path_spar = philips_path / 'hyper' / 'HBCD_HYPER_r5712_Export_WIP_HYPER_5_2_raw_ref.SPAR'


def test_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'philips',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           str(svs_path_sdat),
                           str(svs_path_spar)])

    img_t = read_nifti_mrs(tmp_path / 'svs.nii.gz')

    assert img_t.shape == (1, 1, 1, 2048)
    assert np.iscomplexobj(img_t.dataobj)
    assert 1 / img_t.header['pixdim'][4] == 5000.0

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.759464
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == svs_path_sdat.name
    assert hdr_ext['SoftwareVersions'] == '5.5.2 ; .5.2 ;'

    # Check no hanging singleton dimension
    nmrs_obj = NIFTI_MRS(tmp_path / 'svs.nii.gz')
    assert nmrs_obj.shape == (1, 1, 1, 2048)

    subprocess.check_call([
        'spec2nii', 'philips',
        '-t', 'DIM_DYN',
        '-f', 'svs_singleton',
        '-o', tmp_path,
        '-j',
        str(svs_path_sdat),
        str(svs_path_spar)])

    img_t = read_nifti_mrs(tmp_path / 'svs_singleton.nii.gz')
    assert img_t.shape == (1, 1, 1, 2048)

    nmrs_obj = NIFTI_MRS(tmp_path / 'svs_singleton.nii.gz')
    assert nmrs_obj.shape == (1, 1, 1, 2048, 1)

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())
    assert hdr_ext['dim_5'] == 'DIM_DYN'


def test_multiavg_svs(tmp_path):

    subprocess.check_call(['spec2nii', 'philips',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           str(svs_ma_path_sdat),
                           str(svs_ma_path_spar)])

    img_t = read_nifti_mrs(tmp_path / 'svs.nii.gz')

    assert img_t.shape == (1, 1, 1, 2048, 10)
    assert np.iscomplexobj(img_t.dataobj)

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['dim_5'] == 'DIM_DYN'

    # Check override tags
    subprocess.check_call([
        'spec2nii', 'philips',
        '-t', 'DIM_USER_0',
        '-f', 'svs_2',
        '-o', tmp_path,
        '-j',
        str(svs_ma_path_sdat),
        str(svs_ma_path_spar)])

    img_t = read_nifti_mrs(tmp_path / 'svs_2.nii.gz')
    assert img_t.shape == (1, 1, 1, 2048, 10)

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())
    assert hdr_ext['dim_5'] == 'DIM_USER_0'


def test_svs_edit(tmp_path):

    subprocess.check_call(['spec2nii', 'philips',
                           '-f', 'edit',
                           '-s', '4', '80',
                           '-t', 'DIM_EDIT', 'DIM_DYN',
                           '-o', tmp_path,
                           '-j',
                           str(svs_edit_path_sdat),
                           str(svs_edit_path_spar)])

    img_t = read_nifti_mrs(tmp_path / 'edit.nii.gz')

    assert img_t.shape == (1, 1, 1, 2048, 4, 80)
    assert np.iscomplexobj(img_t.dataobj)
    assert 1 / img_t.header['pixdim'][4] == 2000.0

    hdr_ext_codes = img_t.header.extensions.get_codes()
    hdr_ext = json.loads(img_t.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.768616
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == svs_edit_path_sdat.name
    assert hdr_ext['dim_5'] == 'DIM_EDIT'
    assert hdr_ext['dim_6'] == 'DIM_DYN'


def test_svs_hyper(tmp_path):

    subprocess.check_call(['spec2nii', 'philips',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           str(hyper_path_sdat),
                           str(hyper_path_spar)])

    img_1 = read_nifti_mrs(tmp_path / 'svs_hyper_short_te.nii.gz')
    img_2 = read_nifti_mrs(tmp_path / 'svs_hyper_edited.nii.gz')

    assert img_1.shape == (1, 1, 1, 2048, 32)
    assert np.iscomplexobj(img_1.dataobj)
    assert 1 / img_1.header['pixdim'][4] == 2000.0

    assert img_2.shape == (1, 1, 1, 2048, 4, 56)
    assert np.iscomplexobj(img_2.dataobj)
    assert 1 / img_2.header['pixdim'][4] == 2000.0

    hdr_ext_codes = img_1.header.extensions.get_codes()
    hdr_ext = json.loads(img_1.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.74876
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == hyper_path_sdat.name
    assert hdr_ext['dim_5'] == 'DIM_DYN'

    hdr_ext_codes = img_2.header.extensions.get_codes()
    hdr_ext = json.loads(img_2.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['dim_5'] == 'DIM_EDIT'
    assert hdr_ext['dim_5_header'] == {'EditCondition': ['A', 'B', 'C', 'D']}
    assert hdr_ext['dim_6'] == 'DIM_DYN'


def test_svs_hyper_ref(tmp_path):

    subprocess.check_call(['spec2nii', 'philips',
                           '-f', 'svs',
                           '-o', tmp_path,
                           '-j',
                           str(hyper_ref_path_sdat),
                           str(hyper_ref_path_spar)])

    assert (tmp_path / 'svs_hyper_ref_short_te.nii.gz').is_file()
    assert (tmp_path / 'svs_hyper_ref_edited.nii.gz').is_file()

    img_1 = read_nifti_mrs(tmp_path / 'svs_hyper_ref_short_te.nii.gz')
    img_2 = read_nifti_mrs(tmp_path / 'svs_hyper_ref_edited.nii.gz')

    assert img_1.shape == (1, 1, 1, 2048, 4)
    assert np.iscomplexobj(img_1.dataobj)
    assert 1 / img_1.header['pixdim'][4] == 2000.0

    assert img_2.shape == (1, 1, 1, 2048, 4)
    assert np.iscomplexobj(img_2.dataobj)
    assert 1 / img_2.header['pixdim'][4] == 2000.0

    hdr_ext_codes = img_1.header.extensions.get_codes()
    hdr_ext = json.loads(img_1.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['SpectrometerFrequency'][0] == 127.74876
    assert hdr_ext['ResonantNucleus'][0] == '1H'
    assert hdr_ext['OriginalFile'][0] == hyper_ref_path_sdat.name
    assert hdr_ext['dim_5'] == 'DIM_DYN'

    hdr_ext_codes = img_2.header.extensions.get_codes()
    hdr_ext = json.loads(img_2.header.extensions[hdr_ext_codes.index(44)].get_content())

    assert hdr_ext['dim_5'] == 'DIM_DYN'



"""spec2nii module containing functions specific to interpreting Philips formats
Author: William Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2020 University of Oxford
"""
from datetime import datetime
from ast import literal_eval

import numpy as np
from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

from spec2nii.nifti_orientation import NIFTIOrient, calc_affine
from spec2nii import __version__ as spec2nii_ver

default_tag_order = ['DIM_DYN', 'DIM_EDIT', 'DIM_USER_0']


def read_sdat_spar_pair(sdat_file, spar_file, shape=None, tags=None, fileout=None, special=None):
    """Read and convert SPAR/SDAT pairs from Philips scanners

    :param sdat_file: Path to .SDAT file
    :type sdat_file: pathlib.Path
    :param spar_file: Path to .SPAR file
    :type spar_file: pathlib.Path
    :param shape: List of dimension shapes to reshape array, defaults to None
    :type shape: list of ints, optional
    :param tags: List of higher dimension tags, defaults to None
    :type tags: List of strings, optional
    :param fileout: File name string passed by user, defaults to None/stem of file
    :type fileout: str, optional
    :param special: Identifier for special-cased sequence, defaults to None
    :type special: str, optional
    :return: List of NIFTI-MRS objects
    :rtype: list of nifti_mrs.nifti_mrs.NIFTI_MRS
    :return: List of file name parts
    :rtype: list of strings
    """
    spar_params = read_spar(spar_file)
    data = read_sdat(sdat_file,
                     spar_params['samples'],
                     spar_params['rows'])

    if data.ndim < 4:
        data = data.reshape((1, 1, 1) + data.shape)

    if shape is not None:
        it_shape = data.shape[:4]
        data = data.reshape(it_shape + tuple(shape), order='F')

    # Move to right handed frame
    data = data.conj()

    # Dwelltime
    dwelltime = 1.0 / float(spar_params["sample_frequency"])

    # Meta
    meta = spar_to_nmrs_hdrext(spar_params)
    meta.set_standard_def('OriginalFile', [sdat_file.name])

    # Sort dimension tags
    # First ensure user defined tags has length of 3.
    while len(tags) < 3:
        tags.append(None)

    # Set user defined tags, or the default if the dimension exists
    # and is larger than 1
    for idx, (tag, default) in enumerate(zip(tags, default_tag_order)):
        npdim = idx + 4
        if tag is not None:
            meta.set_dim_info(idx, tag)
        elif data.ndim > npdim and data.shape[npdim] > 1:
            meta.set_dim_info(idx, default)

    # Orientation
    if spar_params["volume_selection_enable"] == "yes":
        affine = _philips_orientation(spar_params)
    else:
        # Use default affine
        affine = np.diag(np.array([10000, 10000, 10000, 1]))
    orientation = NIFTIOrient(affine)

    # name of output
    if fileout is not None:
        mainStr = fileout
    else:
        mainStr = sdat_file.stem

    # Special cases
    if (spar_params['scan_id'].lower() == 'hyper'
        and '_ref' in sdat_file.stem)\
        or (special is not None and
            special.lower() == 'hyper-ref'):
        return _special_case_hyper_ref(data, dwelltime, meta, orientation.Q44, mainStr)
    elif spar_params['scan_id'].lower() == 'hyper'\
            or (special is not None and
                special.lower() == 'hyper'):
        return _special_case_hyper(data, dwelltime, meta, orientation.Q44, mainStr)
    else:
        # Normal case
        return [gen_nifti_mrs_hdr_ext(data, dwelltime, meta, orientation.Q44, no_conj=True), ], \
               [mainStr, ]


def read_spar(filename):
    '''Read the .spar file.
    :param filename: file path

    :return: dict of parameters read from spar file
    :rtype: dict
    '''

    parameter_dict = {}
    with open(filename, 'r') as f:
        for line in f:
            # ignore comments (!) and empty lines
            if line == "\n" or line.startswith("!"):
                continue

            # Handle
            key, value = map(str.strip, line.split(":", 1))
            try:
                val = literal_eval(value)
            except (ValueError, SyntaxError):
                if value == '':
                    val = None
                else:
                    val = value

            parameter_dict.update({key: val})

    return parameter_dict


def read_sdat(filename, samples, rows):
    '''Read the .sdat file.
    :param filename: File path
    :param samples: Number of spectral points
    :param rows: Number of rows of data
    '''
    with open(filename, 'rb') as f:
        raw = f.read()

    floats = _vax_to_ieee_single_float(raw)
    data_iter = iter(floats)
    complex_iter = (complex(r, i) for r, i in zip(data_iter, data_iter))
    raw_data = np.fromiter(complex_iter, "complex64")
    raw_data = np.reshape(raw_data, (rows, samples)).T.squeeze()

    return raw_data


def _philips_orientation(params):
    '''Calculate the orientation affine from the spar parameters.'''

    angle_lr = params['lr_angulation']
    angle_ap = params['ap_angulation']
    angle_hf = params['cc_angulation']
    angles = [-angle_lr, -angle_ap, angle_hf]

    dim_lr = params['lr_size']
    dim_ap = params['ap_size']
    dim_hf = params['cc_size']
    dimensions = [dim_lr, dim_ap, dim_hf]

    shift_lr = params['lr_off_center']
    shift_ap = params['ap_off_center']
    shift_hf = params['cc_off_center']
    shift = [-shift_lr, -shift_ap, shift_hf]

    return calc_affine(angles, dimensions, shift)


def spar_to_nmrs_hdrext(spar_dict):
    """ Extract information from the dict of keys read from the spar file to insert into the json header ext.

    :param dict spar_dict: key-value pairs read from spar file
    :return: NIfTI MRS hdr ext object.
    """

    # Extract required metadata and create hdr_ext object
    cf = float(spar_dict["synthesizer_frequency"]) / 1E6
    obj = Hdr_Ext(
        cf,
        spar_dict["nucleus"])

    def set_standard_def(nifti_mrs_key, location, key, cast=None):
        if key in location\
                and location[key] is not None:
            if cast is not None:
                obj.set_standard_def(nifti_mrs_key, cast(location[key]))
            else:
                obj.set_standard_def(nifti_mrs_key, location[key])

    # # 5.1 MRS specific Tags
    # 'EchoTime'
    obj.set_standard_def('EchoTime', float(spar_dict['echo_time']) * 1E-3)
    # 'RepetitionTime'
    obj.set_standard_def('RepetitionTime', float(spar_dict['repetition_time'] / 1E3))
    # 'InversionTime'
    if spar_dict['spectrum_inversion_time'] > 0:
        obj.set_standard_def('InversionTime', float(spar_dict['spectrum_inversion_time']))
    # 'MixingTime'
    # 'ExcitationFlipAngle'
    # 'TxOffset'
    obj.set_standard_def('TxOffset', float(spar_dict['offset_frequency'] / cf))
    # 'VOI'
    # 'WaterSuppressed'
    # No apparent parameter stored in the SPAR info.
    # 'WaterSuppressionType'
    # 'SequenceTriggered'
    # # 5.2 Scanner information
    # 'Manufacturer'
    obj.set_standard_def('Manufacturer', 'Philips')
    # 'ManufacturersModelName'
    # 'DeviceSerialNumber'
    # 'SoftwareVersions'
    set_standard_def('SoftwareVersions', spar_dict, 'equipment_sw_verions')
    # 'InstitutionName'
    # 'InstitutionAddress'
    # 'TxCoil'
    # 'RxCoil'
    # # 5.3 Sequence information
    # 'SequenceName'
    # 'ProtocolName'
    set_standard_def('ProtocolName', spar_dict, 'scan_id')
    # # 5.4 Sequence information
    # 'PatientPosition'
    try:
        obj.set_standard_def('PatientPosition', spar_dict['patient_position'] + ' ' + spar_dict['patient_orientation'])
    except AttributeError:
        pass
    # 'PatientName'
    set_standard_def('PatientName', spar_dict, 'patient_name')
    # 'PatientID'
    # 'PatientWeight'
    # 'PatientDoB'
    set_standard_def('PatientDoB', spar_dict, 'patient_birth_date')
    # 'PatientSex'
    # # 5.5 Provenance and conversion metadata
    # 'ConversionMethod'
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)
    # 'OriginalFile'
    # Set elsewhere
    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    return obj


# From VESPA - BSD license.
def _vax_to_ieee_single_float(data):
    """Converts a float in Vax format to IEEE format.

    data should be a single string of chars that have been read in from
    a binary file. These will be processed 4 at a time into float values.
    Thus the total number of byte/chars in the string should be divisible
    by 4.

    Based on VAX data organization in a byte file, we need to do a bunch of
    bitwise operations to separate out the numbers that correspond to the
    sign, the exponent and the fraction portions of this floating point
    number

    role :      S        EEEEEEEE      FFFFFFF      FFFFFFFF      FFFFFFFF
    bits :      1        2      9      10                               32
    bytes :     byte2           byte1               byte4         byte3

    This is taken from the VESPA project source code under a BSD licence.
    """
    f = []
    nfloat = int(len(data) / 4)
    for i in range(nfloat):

        byte2 = data[0 + i * 4]
        byte1 = data[1 + i * 4]
        byte4 = data[2 + i * 4]
        byte3 = data[3 + i * 4]

        # hex 0x80 = binary mask 10000000
        # hex 0x7f = binary mask 01111111

        sign = (byte1 & 0x80) >> 7
        expon = ((byte1 & 0x7f) << 1) + ((byte2 & 0x80) >> 7)
        fract = ((byte2 & 0x7f) << 16) + (byte3 << 8) + byte4

        if sign == 0:
            sign_mult = 1.0
        else:
            sign_mult = -1.0

        if 0 < expon:
            # note 16777216.0 == 2^24
            val = sign_mult * (0.5 + (fract / 16777216.0)) * pow(2.0, expon - 128.0)
            f.append(val)
        elif expon == sign == 0:
            f.append(0)
        else:
            f.append(0)
            # may want to raise an exception here ...

    return f


def _special_case_hyper(data, dwelltime, meta, orientation, fout_str):
    """Special case handling for the HYPER sequence (main water suppressed component).

    Data is split into short TE and editing (4 conditions) blocks.

    :param data: Combined short TE and editing condition data
    :type data: np.ndarray
    :param dwelltime: Spectral dwelltime
    :type dwelltime: float
    :param meta: raw meta object extracted from data
    :type meta: nifti_mrs.hdr_ext.Hdr_Ext
    :param orientation: Orientation object
    :type orientation: spec2nii.nifti_orientation.NIFTIOrient
    :param fout_str: Base of output string
    :type fout_str: str
    :return: 2 x Nifti-MRS objects, one short TE, one editing
    :rtype: tuple of nifti_mrs.nifti_mrs.NIFTI_MRS
    :return: 2 x Output names
    :rtype: tuple of str

    Notes:
      - (AG 03/22/2023) Updated to handle incomplete data.

    """
    # Reorganise the data. This unfortunately makes hardcoded assumptions about the size of each part.
    data_short_te = data[:, :, :, :, :32]
    data_edited = data[:, :, :, :, 32:]

    if data_edited.shape[-1] % 4 != 0:                                                      # Handle Incomplete Data
        old_num_avgs = data_edited.shape[-1]                                                # Old Number of Averages
        new_num_avgs = (data_edited.shape[-1] // 4) * 4                                     # Complete Sets of 4
        data_edited = data_edited[..., :new_num_avgs]                                       # Only Keep Complete Sets
        print(f'Correcting - Incomplete Averages {old_num_avgs} --> {new_num_avgs}    Corrected**')

    data_edited = data_edited.T.reshape((-1, 4, data.shape[3], 1, 1, 1)).T

    meta_short_te = meta.copy()
    meta_short_te.set_standard_def("WaterSuppressed", True)
    meta_short_te.set_standard_def("EchoTime", 0.035)

    meta_edited = meta.copy()

    edit_pulse_1 = 1.9
    edit_pulse_2 = 4.58
    edit_pulse_off = 4.18
    dim_info = "HERCULES j-difference editing, four conditions"
    dim_header = {"EditCondition": ["A", "B", "C", "D"]}
    edit_pulse_val = {
        "A": {"PulseOffset": [edit_pulse_1, edit_pulse_2], "PulseDuration": 0.02},
        "B": {"PulseOffset": [edit_pulse_1, edit_pulse_off], "PulseDuration": 0.02},
        "C": {"PulseOffset": edit_pulse_2, "PulseDuration": 0.02},
        "D": {"PulseOffset": edit_pulse_off, "PulseDuration": 0.02}}

    meta_edited.set_dim_info(
        0,
        'DIM_EDIT',
        dim_info,
        dim_header)

    meta_edited.set_dim_info(1, 'DIM_DYN')
    meta_edited.set_standard_def("EditPulse", edit_pulse_val)
    meta_edited.set_standard_def("WaterSuppressed", True)
    meta_edited.set_standard_def("EchoTime", 0.08)

    return [gen_nifti_mrs_hdr_ext(data_short_te, dwelltime, meta_short_te, orientation, no_conj=True),
            gen_nifti_mrs_hdr_ext(data_edited, dwelltime, meta_edited, orientation, no_conj=True)], \
           [fout_str + '_hyper_short_te', fout_str + '_hyper_edited']


def _special_case_hyper_ref(data, dwelltime, meta, orientation, fout_str):
    """Special case handling for the HYPER sequence (water reference component).

    Contains 8 references (every 32 scans from index 0). Starts with TE = 80 and alternates to TE = 35.
    Output two ntime x 4 (DIM_DYN) shape files to match main acquisition

    :param data: Combined short TE and editing-TE data
    :type data: numpy.ndarray
    :param dwelltime: Spectral dwelltime
    :type dwelltime: float
    :param meta: raw meta object extracted from data
    :type meta: nifti_mrs.hdr_ext.Hdr_Ext
    :param orientation: Orientation object
    :type orientation: spec2nii.nifti_orientation.NIFTIOrient
    :param fout_str: Base of output string
    :type fout_str: str
    :return: 2 x Nifti-MRS objects, one short TE, one editing-TE
    :rtype: tuple of nifti_mrs.nifti_mrs.NIFTI_MRS
    :return: 2 x Output names
    :rtype: tuple of str
    """
    # Reorganise the data. This unfortunately makes hardcoded assumptions about the size of each part.
    data_ref = data[:, :, :, :, ::32]
    data_edited = data_ref[:, :, :, :, 0::2]
    data_short_te = data_ref[:, :, :, :, 1::2]

    meta_short_te = meta.copy()
    meta_short_te.set_dim_info(0, 'DIM_DYN')
    meta_short_te.set_standard_def("WaterSuppressed", False)
    meta_short_te.set_standard_def("EchoTime", 0.035)

    meta_edited = meta.copy()
    meta_edited.set_dim_info(0, 'DIM_DYN')
    meta_edited.set_standard_def("WaterSuppressed", False)
    meta_short_te.set_standard_def("EchoTime", 0.08)

    return [gen_nifti_mrs_hdr_ext(data_short_te, dwelltime, meta_short_te, orientation, no_conj=True),
            gen_nifti_mrs_hdr_ext(data_edited, dwelltime, meta_edited, orientation, no_conj=True)], \
           [fout_str + '_hyper_ref_short_te', fout_str + '_hyper_ref_edited']



"""spec2nii module containing functions specific to interpreting Bruker formats
Dependent on the brukerapi package developed by Tomas Psorn.
https://github.com/isi-nmr/brukerapi-python

Author: Tomas Psorn <tomaspsorn@isibrno.cz>
        Will Clarke <william.clarke@ndcn.ox.ac.uk>
Copyright (C) 2021 Institute of Scientific Instruments of the CAS, v. v. i.
"""
import os
import pkg_resources
import warnings
from datetime import datetime

import numpy as np

from brukerapi.dataset import Dataset
from brukerapi.folders import Folder
from brukerapi.mergers import FrameGroupMerger
from brukerapi.exceptions import FilterEvalFalse

from nifti_mrs.create_nmrs import gen_nifti_mrs_hdr_ext
from nifti_mrs.hdr_ext import Hdr_Ext

from spec2nii.nifti_orientation import NIFTIOrient
from spec2nii import __version__ as spec2nii_ver

# Default dimension assignments.
fid_dimension_defaults = {
    'repetition': "DIM_DYN",
    'channel': "DIM_COIL"}


def read_bruker(args):
    """

    :param args:
    :return list imageOut:
    :return list fileoutNames:
    """
    imageOut = []
    fileoutNames = []

    # for all Bruker datasets compliant all queries
    for data, orientation, dwelltime, meta, name in yield_bruker(args):
        imageOut.append(
            gen_nifti_mrs_hdr_ext(
                data,
                dwelltime,
                meta,
                orientation.Q44,
                no_conj=True)
        )
        fileoutNames.append(name)

    return imageOut, fileoutNames


def yield_bruker(args):
    """

    If the path specified by args.file is:

    1/ Bruker dataset file (2dseq) - function yields its data and properties of the dataset
    2/ Directory - function yields data and properties and data of all datasets compliant to the queries

    """
    # get location of the spec2nii Bruker properties configuration file
    bruker_properties_path = pkg_resources.resource_filename('spec2nii', 'bruker_properties.json')
    bruker_fid_override_path = pkg_resources.resource_filename('spec2nii', 'bruker_fid_override.json')

    # get a list of queries to filter datasets
    queries = _get_queries(args)

    # case of Bruker dataset
    if os.path.isfile(args.file):
        d = Dataset(
            args.file,
            property_files=[bruker_fid_override_path, bruker_properties_path],
            parameter_files=['method'])
        try:
            d.query(queries)
        except FilterEvalFalse:
            raise ValueError(f'Bruker dataset {d.path} is not suitable for conversion to mrs_nifti')
        yield from _proc_dataset(d, args)

    # case of folder containing Bruker datasets
    elif os.path.isdir(args.file):

        # process individual datasets
        for dataset in Folder(args.file, dataset_state={
            "parameter_files": ['method'],
            "property_files": [bruker_properties_path]
        }).get_dataset_list_rec():
            with dataset as d:
                try:
                    d.query(queries)
                except FilterEvalFalse:
                    continue
                yield from _proc_dataset(d, args)


def _get_queries(args):
    """
    Returns a list of queries for filtering out only spectroscopic 2dseq datasets with a complex frame group

    """
    if args.mode == '2DSEQ':
        queries = ["@type=='2dseq'", "@is_spectroscopy==True", "@is_complex==True"]
    elif args.mode == 'FID':
        queries = ["@type=='fid'", "@is_spectroscopy==True"]
    return queries + args.query


def _proc_dataset(d, args):
    """
    Yield data and properties of a single dataset

    """
    # merge 2dseq complex frame group if present
    if d.is_complex and d.type == '2dseq':
        d = FrameGroupMerger().merge(d, 'FG_COMPLEX')

    # prepare the data array
    if d.is_svs:
        data = _prep_data_svs(d)
    elif d.is_mrsi:
        data = _prep_data_mrsi(d)
    else:
        data = d.data

    # get properties
    properties = d.to_dict()

    # Orientation information
    if d.type == 'fid':
        orientation = NIFTIOrient(_fid_affine_from_params(d))
    else:
        orientation = NIFTIOrient(np.reshape(np.array(properties['affine']), (4, 4)))

    # Meta data
    if d.type == 'fid':
        meta = _fid_meta(d, dump=args.dump_headers)
    else:
        meta = _2dseq_meta(d, dump=args.dump_headers)

    # Dwelltime - to do resolve this factor of 2 issue
    if d.type == 'fid':
        dwelltime = d.dwell_s * 2
    else:
        dwelltime = d.dwell_s * 2

    if args.fileout:
        name = args.fileout + '_' + d.id.rstrip('_')
    else:
        name = d.id.rstrip('_')

    yield data, orientation, dwelltime, meta, name


def _prep_data_svs(d):
    """
    Push the spectral dimension of the data array to the 3rd position for SVS data

    It is possible to use tuple as an axis argument of the expand_dims function since numpy>=1.18.0,
    we decided to use this triple call to avoid limiting numpy versions

    """
    data = d.data
    if d.type == 'fid':
        # Remove points acquired before echo
        data = data[d.points_prior_to_echo:, ...]

        # fid data appears to need to be conjugated for NIFTI-MRS convention
        data = data.conj()

    data = np.expand_dims(data, axis=0)
    data = np.expand_dims(data, axis=0)
    data = np.expand_dims(data, axis=0)
    return data


def _prep_data_mrsi(d):
    """
    Push the spectral dimension of the data array to the 3rd position for CSI data

    """
    data = d.data
    if d.type == 'fid':
        # Remove points acquired before echo
        data = data[d.points_prior_to_echo:, ...]

        # fid data appears to need to be conjugated for NIFTI-MRS convention
        data = data.conj()

    # push the spectral dimension to position 2
    data = np.moveaxis(data, 0, 2)
    # add empty dimensions to push the spectral dimension to the 3rd index
    data = np.expand_dims(data, axis=2)
    return data


def _fid_affine_from_params(d):
    """ First attempt to create 4x4 affine from fid headers"""
    warnings.warn('The orientation of bruker fid data is mostly untested.')

    orientation = np.squeeze(d.parameters['method']['PVM_VoxArrGradOrient'].value)
    shift = np.squeeze(d.parameters['method']['PVM_VoxArrPosition'].value)
    # csshift = np.squeeze(d.parameters['method']['PVM_VoxArrCSDisplacement'].value)
    # shift += csshift
    size = np.squeeze(d.parameters['method']['PVM_VoxArrSize'].value)
    affine = np.zeros((4, 4))
    affine[3, 3] = 1
    reorder = [1, 2, 0]  # [0, 1, 2] *[0, 2, 1]* [1, 2, 0]
    affine[:3, :3] = orientation[reorder, :].T * size[reorder]
    affine[:3, 3] = shift[reorder]

    return affine


def _2dseq_meta(d, dump=False):
    """ Extract information from method and acqp file into hdr_ext.

    :param d: Dataset
    :return: NIfTI MRS hdr ext object.
    """

    # Extract required metadata and create hdr_ext object
    cf = d.SpectrometerFrequency
    obj = Hdr_Ext(
        cf,
        d.ResonantNucleus)

    # # 5.1 MRS specific Tags
    # 'EchoTime'
    if hasattr(d, 'TE'):
        obj.set_standard_def('EchoTime', float(d.TE * 1E-3))
    elif hasattr(d, 'method_TE'):
        obj.set_standard_def('EchoTime', float(d.method_TE * 1E-3))
    # 'RepetitionTime'
    if hasattr(d, 'TR'):
        obj.set_standard_def('RepetitionTime', float(d.TR / 1E3))
    elif hasattr(d, 'method_TR'):
        obj.set_standard_def('RepetitionTime', float(d.method_TR / 1E3))
    # 'InversionTime'
    # 'MixingTime'
    # 'ExcitationFlipAngle'
    # 'TxOffset'
    # Bit of a guess, not sure of units.
    obj.set_standard_def('TxOffset', float(d.working_offset[0]))
    # 'VOI'
    # 'WaterSuppressed'
    # No apparent parameter stored in the SPAR info.
    # 'WaterSuppressionType'
    # 'SequenceTriggered'
    # # 5.2 Scanner information
    # 'Manufacturer'
    obj.set_standard_def('Manufacturer', 'Bruker')
    # 'ManufacturersModelName'
    # 'DeviceSerialNumber'
    # 'SoftwareVersions'
    obj.set_standard_def('SoftwareVersions', d.PV_version)
    # 'InstitutionName'
    # 'InstitutionAddress'
    # 'TxCoil'
    # 'RxCoil'
    # # 5.3 Sequence information
    # 'SequenceName'
    obj.set_standard_def('SequenceName', d.method_desc)
    # 'ProtocolName'
    # # 5.4 Sequence information
    # 'PatientPosition'
    # 'PatientName'
    obj.set_standard_def('PatientName', d.subj_id)
    # 'PatientID'
    # 'PatientWeight'
    # 'PatientDoB'
    # 'PatientSex'
    # # 5.5 Provenance and conversion metadata
    # 'ConversionMethod'
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)
    # 'OriginalFile'
    obj.set_standard_def('OriginalFile', [str(d.path), ])
    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    # Stuff full headers into user fields
    if dump:
        for hdr_file in d.parameters:
            obj.set_user_def(key=hdr_file,
                             doc=f'Bruker {hdr_file} file.',
                             value=d.parameters[hdr_file].to_dict())

    # Tags
    unknown_count = 0
    for ddx, dim in enumerate(d.dim_type[1:]):
        if dim in fid_dimension_defaults:
            obj.set_dim_info(ddx, fid_dimension_defaults[dim])
        else:
            obj.set_dim_info(ddx, f'DIM_USER_{unknown_count}')
            unknown_count += 1

    return obj


def _fid_meta(d, dump=False):
    """ Extract information from method and acqp file into hdr_ext.

    :param d: Dataset
    :return: NIfTI MRS hdr ext object.
    """

    # Extract required metadata and create hdr_ext object
    cf = d.SpectrometerFrequency
    obj = Hdr_Ext(
        cf,
        d.ResonantNucleus)

    # # 5.1 MRS specific Tags
    # 'EchoTime'
    obj.set_standard_def('EchoTime', float(d.TE * 1E-3))
    # 'RepetitionTime'
    obj.set_standard_def('RepetitionTime', float(d.TR / 1E3))
    # 'InversionTime'
    # 'MixingTime'
    # 'ExcitationFlipAngle'
    # 'TxOffset'
    # Bit of a guess, not sure of units.
    obj.set_standard_def('TxOffset', float(d.working_offset[0]))
    # 'VOI'
    # 'WaterSuppressed'
    # No apparent parameter stored in the SPAR info.
    # 'WaterSuppressionType'
    # 'SequenceTriggered'
    # # 5.2 Scanner information
    # 'Manufacturer'
    obj.set_standard_def('Manufacturer', 'Bruker')
    # 'ManufacturersModelName'
    # 'DeviceSerialNumber'
    # 'SoftwareVersions'
    obj.set_standard_def('SoftwareVersions', d.PV_version)
    # 'InstitutionName'
    # 'InstitutionAddress'
    # 'TxCoil'
    # 'RxCoil'
    # # 5.3 Sequence information
    # 'SequenceName'
    obj.set_standard_def('SequenceName', d.method_desc)
    # 'ProtocolName'
    # # 5.4 Sequence information
    # 'PatientPosition'
    # 'PatientName'
    obj.set_standard_def('PatientName', d.subj_id)
    # 'PatientID'
    # 'PatientWeight'
    # 'PatientDoB'
    # 'PatientSex'
    # # 5.5 Provenance and conversion metadata
    # 'ConversionMethod'
    obj.set_standard_def('ConversionMethod', f'spec2nii v{spec2nii_ver}')
    # 'ConversionTime'
    conversion_time = datetime.now().isoformat(sep='T', timespec='milliseconds')
    obj.set_standard_def('ConversionTime', conversion_time)
    # 'OriginalFile'
    obj.set_standard_def('OriginalFile', [str(d.path), ])
    # # 5.6 Spatial information
    # 'kSpace'
    obj.set_standard_def('kSpace', [False, False, False])

    # Stuff full headers into user fields
    if dump:
        for hdr_file in d.parameters:
            obj.set_user_def(key=hdr_file,
                             doc=f'Bruker {hdr_file} file.',
                             value=d.parameters[hdr_file].to_dict())

    # Tags
    unknown_count = 0
    for ddx, dim in enumerate(d.dim_type[1:]):
        if dim in fid_dimension_defaults:
            obj.set_dim_info(ddx, fid_dimension_defaults[dim])
        else:
            obj.set_dim_info(ddx, f'DIM_USER_{unknown_count}')
            unknown_count += 1

    return obj
